/* FILE: mmdispcmds.cc                 -*-Mode: c++-*-
 *
 * Tcl commands implemented in C/C++ and supported by shell mmdispsh
 *
 */

#include <math.h>	/* fmod(), ... */
#include <stdlib.h>	/* atoi(), ... */
#include <string.h>	/* strncpy(), ... */

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "mmdispcmds.h"
#include "display.h"	/* DisplayFrame, CoordinateSystem */
#include "bitmap.h"

/* End includes */

// The TclObj interfaces are available in Tcl later than 8.0; earlier
// than that we must use the string interface.  However, one can set
// the macro USE_TCLOBJS to 0 to force the use of the string interface.
// This may be useful for debugging purposes.
#ifndef USE_TCLOBJS
# if TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0)
#  define USE_TCLOBJS 1
# else
#  define USE_TCLOBJS 0
# endif
#endif

// A couple of forward declarations.  These shouldn't be accessed
// from outside this file.  The proper way to insure this is to
// use the "unnamed" namespace.  But for backwards compatibility
// with old compilers, we aren't using namespaces.  The old way
// to do this is to declare these functions "static", but that use
// of "static" is deprecated.  For now, just punt and put them in
// the global namespace.
REAL4m SetZoom(REAL4m zoom);
REAL4m SetZoom(Tcl_Interp *interp,REAL4m width,REAL4m height);

#define MY_MESH_ARRAY_SIZE 2
static Vf_Mesh* myMeshArray[MY_MESH_ARRAY_SIZE];
static int activeMeshId = 0;
static Vf_Mesh *myMesh;
static DisplayFrame myFrame;

#define MY_BUF_SIZE 8192
static CHAR buf[MY_BUF_SIZE];

static REAL4m CoordsToAngle(CoordinateSystem coords)
{
  REAL4m rotang=0.;
  switch(coords) {
  case CS_CALCULATION_STANDARD: // Safety
  case CS_DISPLAY_STANDARD:  rotang=0.;   break;
  case CS_DISPLAY_ROT_90:    rotang=90.;  break;
  case CS_DISPLAY_ROT_180:   rotang=180.; break;
  case CS_DISPLAY_ROT_270:   rotang=270.; break;
  case CS_ILLEGAL:
  default:
    PlainError(1,"Unrecognized or illegal coordinate system (%d)"
	       " detected in CoordsToAngle() (File tkcmds.cc)\n",
	       int(coords));
    break; // Shouldn't get to here
  }
  return rotang;
}

static CoordinateSystem AngleToCoords(REAL4m angle_in_degrees)
{ // Rounds to nearest supported coordinate system rotation
  CoordinateSystem coords=CS_ILLEGAL;
  int quad=int(OC_ROUND(fmod(double(angle_in_degrees),
			     double(360.))/90.));
  if(quad<0) quad+=4;  if(quad>3) quad-=4;
  switch(quad) {
  case 0:  coords=CS_DISPLAY_STANDARD;  break;
  case 1:  coords=CS_DISPLAY_ROT_90;    break;
  case 2:  coords=CS_DISPLAY_ROT_180;   break;
  case 3:  coords=CS_DISPLAY_ROT_270;   break;
  default:
    PlainError(1,"Programming error: Illegal coordinate system (quad=%d)"
	       " occurred in AngleToCoords() (File tkcmds.cc)\n",quad);
    break; // Shouldn't get to here
  }
  return coords;
}

int ReportActiveMesh(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s\"", argv[0]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),"%d",activeMeshId);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}


int SelectActiveMesh(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
		"SelectActiveMesh must be called with 1 argument:"
		" Mesh Id to make active (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int meshId = atoi(argv[1]);
  if(meshId<0 || meshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		meshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int oldId = activeMeshId;
  activeMeshId = meshId;
  myMesh = myMeshArray[activeMeshId];

  myFrame.SetMesh(myMesh);

  Oc_Snprintf(buf,sizeof(buf),"%d",oldId);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int FreeMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s meshid\"", argv[0]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int meshId = atoi(argv[1]);
  if(meshId<0 || meshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		meshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  if (myMesh==myMeshArray[meshId]) myMesh=NULL; //Safety
  delete myMeshArray[meshId];
  myMeshArray[meshId]=new Vf_EmptyMesh();
  if(activeMeshId == meshId) myMesh = myMeshArray[meshId];

  return TCL_OK;
}

int CopyMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<3 || argc>7) {
    Oc_Snprintf(buf,sizeof(buf),
		"CopyMesh must be called with 2-6 arguments: "
		"original mesh id, new mesh id, "
		"subsample, flip string, clipbox, clip_range"
		" (%d arguments passed)",
		argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int origmeshId = atoi(argv[1]);
  if(origmeshId<0 || origmeshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		origmeshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int newmeshId = atoi(argv[2]);
  if(newmeshId<0 || newmeshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid mesh id request: %d; should be between 0 and %d",
		newmeshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  REAL8m subsample = 0.;
  if(argc>3) subsample = Nb_Atof(argv[3]);
  const char* flipstr = "x:y:z";
  if(argc>4) flipstr = argv[4];

  // Clipbox
  Nb_BoundingBox<REAL8m> clipbox;
  myMeshArray[origmeshId]->GetPreciseRange(clipbox); // All by default
  if(argc>5 && argv[5][0]!='\0') {
    const char* parsestr=" \t\n:";
    char* buf = new char[strlen(argv[5])+1]; // clipbox corners
    strcpy(buf,argv[5]);
    char* nexttoken=buf;
    char *token1 = Nb_StrSep(&nexttoken,parsestr);
    char *token2 = Nb_StrSep(&nexttoken,parsestr);
    char *token3 = Nb_StrSep(&nexttoken,parsestr);
    char *token4 = Nb_StrSep(&nexttoken,parsestr);
    char *token5 = Nb_StrSep(&nexttoken,parsestr);
    char *token6 = Nb_StrSep(&nexttoken,parsestr);

    Nb_Vec3<REAL8m> minpt,maxpt;
    clipbox.GetExtremes(minpt,maxpt);
    if(token1[0]!='\0' && strcmp("-",token1)!=0) {
      minpt.x = Nb_Atof(token1);
    }
    if(token2[0]!='\0' && strcmp("-",token2)!=0) {
      minpt.y = Nb_Atof(token2);
    }
    if(token3[0]!='\0' && strcmp("-",token3)!=0) {
      minpt.z = Nb_Atof(token3);
    }
    if(token4[0]!='\0' && strcmp("-",token4)!=0) {
      maxpt.x = Nb_Atof(token4);
    }
    if(token5[0]!='\0' && strcmp("-",token5)!=0) {
      maxpt.y = Nb_Atof(token5);
    }
    if(token6[0]!='\0' && strcmp("-",token6)!=0) {
      maxpt.z = Nb_Atof(token6);
    }
    delete[] buf;
    clipbox.Set(minpt,maxpt);
  }

  // Clip range?
  BOOL clip_range = 0;
  if(argc>6 && argv[6][0]!='\0') {
    if(atoi(argv[6])) clip_range = 1;
  }

  // Create new mesh; if the original mesh is regular (i.e.,
  // Vf_GridVec3f or Vf_Empty), then make the new one of the
  // same type.
  Vf_Mesh* newmesh=NULL;
  const char* origtype = myMeshArray[origmeshId]->GetMeshType();
  if(strcmp(origtype,"Vf_EmptyMesh")==0) {
    newmesh = new Vf_EmptyMesh();
  } else if(strcmp(origtype,"Vf_GridVec3f")==0) {
    newmesh = new
      Vf_GridVec3f(*((Vf_GridVec3f*)myMeshArray[origmeshId]),
		   subsample,flipstr,clipbox,clip_range);
  } else {
    newmesh = new
      Vf_GeneralMesh3f(*myMeshArray[origmeshId],subsample,
		       flipstr,clipbox,clip_range);
  }

  // Place new mesh into myMeshArray
  if(myMeshArray[newmeshId]!=NULL) {
    if(myMesh==myMeshArray[newmeshId]) myMesh=NULL; // Safety
    delete myMeshArray[newmeshId];
  }
  myMeshArray[newmeshId] = newmesh;
  if(activeMeshId == newmeshId) myMesh = newmesh;

  return TCL_OK;
}

int DifferenceMesh(ClientData,Tcl_Interp *interp,
		   int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
		"DifferenceMesh must be called with 1 argument: "
		"Mesh Id to subtract from active mesh "
		"(%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int meshId = atoi(argv[1]);
  if(meshId<0 || meshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
	    "Invalid mesh id request: %d; should be between 0 and %d",
	    meshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Difference meshes here
  if((*myMeshArray[activeMeshId]).SubtractMesh(*myMeshArray[meshId])!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Meshes aren't compatible");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

int GetMeshType(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshType must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetMeshType();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

static int IsRegularMesh(Vf_Mesh* mesh)
{
  const char *name=mesh->GetMeshType();
  int regular=0;
  if(strcmp(name,"Vf_GridVec3f")==0 ||
     strcmp(name,"Vf_EmptyMesh")==0) {
    // Known regular meshes
    regular=1;
  } // Otherwise assume irregular
  return regular;
}

int IsRegularMesh(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "IsRegularMesh must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int regular = IsRegularMesh(myMesh);
  Oc_Snprintf(buf,sizeof(buf),"%d",regular);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetVecColor(ClientData,Tcl_Interp *interp,
		int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetVecColor must be called with 1 argument:"
	    " A 3 element list representing a vector"
            " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int carr_size;
  CONST84 char** carr;
  if (Tcl_SplitList(interp,argv[1],&carr_size,&carr)!=TCL_OK) {
    return TCL_ERROR;
  }
  if(carr_size!=3) {
    Oc_Snprintf(buf,sizeof(buf),
		"Input list has %d != 3 elements",carr_size);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Nb_Vec3<REAL4> v(static_cast<REAL4m>(Nb_Atof(carr[0])),
		   static_cast<REAL4m>(Nb_Atof(carr[1])),
		   static_cast<REAL4m>(Nb_Atof(carr[2])));
  Tcl_Free((char *)carr);
  myFrame.GetVecColor(v);
  Oc_AutoBuf ab;
  Tcl_AppendResult(interp,ab(myFrame.GetVecColor(v)),(char *)NULL);
  return TCL_OK;
}


int GetMeshName(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshName must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetName();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshStructureInfo(ClientData,Tcl_Interp *interp,
			 int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshStructureInfo must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m size=myMesh->GetSize();

  Nb_BoundingBox<REAL8> datarange;
  myMesh->GetPreciseDataRange(datarange);
  Nb_Vec3<REAL8m> minpt,maxpt;
  datarange.GetExtremes(minpt,maxpt);

  Nb_BoundingBox<REAL8> meshrange;
  myMesh->GetPreciseRange(meshrange);
  Nb_Vec3<REAL8m> mesh_minpt,mesh_maxpt;
  meshrange.GetExtremes(mesh_minpt,mesh_maxpt);

  REAL8m value_min,value_max;
  const char *meshunit = myMesh->GetMeshUnit();
  myMesh->GetValueMagSpan(value_min,value_max);
  const char *valueunit = myMesh->GetValueUnit();
  if(IsRegularMesh(myMesh)) {
    if(size<1 || strcmp("Vf_GridVec3f",myMesh->GetMeshType())!=0) {
      Oc_Snprintf(buf,sizeof(buf),
		  "Regular mesh\n Mesh size: %d\n",size);
    } else {
      Vf_GridVec3f* regmesh = (Vf_GridVec3f*)myMesh;
      INT4m dimx,dimy,dimz;
      regmesh->GetDimens(dimx,dimy,dimz);
      Oc_Snprintf(buf,sizeof(buf),
		  "Regular mesh\n"
		  " Mesh size: %d\n"
		  " Dimensions: %d %d %d\n"
		  " Value magnitude span: %g to %g (in %s)\n"
		  " Data range: (%g,%g,%g) x (%g,%g,%g) (in %s)\n"
		  " Mesh range: (%g,%g,%g) x (%g,%g,%g) (in %s)",
		  size,
		  dimx,dimy,dimz,
		  value_min,value_max,valueunit,
		  minpt.x,minpt.y,minpt.z,
		  maxpt.x,maxpt.y,maxpt.z,meshunit,
		  mesh_minpt.x,mesh_minpt.y,mesh_minpt.z,
		  mesh_maxpt.x,mesh_maxpt.y,mesh_maxpt.z,meshunit);
    }
  } else {
    Oc_Snprintf(buf,sizeof(buf),
		"Irregular mesh\n"
		" Size: %d\n"
		" Value magnitude span: %g to %g (in %s)\n"
		" Data range: (%g,%g,%g) x (%g,%g,%g) (in %s)\n"
		" Mesh range: (%g,%g,%g) x (%g,%g,%g) (in %s)",
		size,
		value_min,value_max,valueunit,
		minpt.x,minpt.y,minpt.z,
		maxpt.x,maxpt.y,maxpt.z,meshunit,
		mesh_minpt.x,mesh_minpt.y,mesh_minpt.z,
		mesh_maxpt.x,mesh_maxpt.y,mesh_maxpt.z,meshunit);
  }
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshTitle(ClientData,Tcl_Interp *interp,int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshTitle must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *title=myMesh->GetTitle();
  if(title!=NULL) strncpy(buf,title,sizeof(buf));
  else            buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int SetMeshTitle(ClientData,Tcl_Interp *interp,
		 int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
            "SetMeshTitle must be called with 1 argument:"
	    " new_title (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  myMesh->SetTitle(argv[1]);
  return TCL_OK;
}

int GetMeshDescription(ClientData,Tcl_Interp *interp,
		       int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshDesc must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *desc=myMesh->GetDescription();
  if(desc!=NULL) strncpy(buf,desc,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshCellSize(ClientData,Tcl_Interp *interp,
		    int argc,CONST84 char** /* argv */)
{
  // Returns a triple: xstep ystep zstep (in mesh units)
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshCellSize must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_Vec3<REAL4> dim=myMesh->GetApproximateCellDimensions();
  Oc_Snprintf(buf,sizeof(buf),"%.8g %.8g %.8g",dim.x,dim.y,dim.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshCoordinates(ClientData,Tcl_Interp *interp,
		    int argc,CONST84 char** argv)
{ // Import: (x,y,z) in display coordinates
  // Export: (x,y,z) in mesh coordinates
  // The returned results are REAL4 precision.
  Tcl_ResetResult(interp);
  if(argc!=4) {
    Oc_Snprintf(buf,sizeof(buf),
		"GetMeshCellSize must be called with"
		" 3 arguments: x y z, in display coordinates"
		" (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL4 x = static_cast<REAL4>(Nb_Atof(argv[1]));
  REAL4 y = static_cast<REAL4>(Nb_Atof(argv[2]));
  REAL4 z = static_cast<REAL4>(Nb_Atof(argv[3]));
  Nb_Vec3<REAL4> pt(x,y,z);
  myFrame.CoordinatePointTransform(myFrame.GetCoordinates(),
				   CS_CALCULATION_STANDARD,
				   pt);
  Oc_Snprintf(buf,sizeof(buf),"%.8g %.8g %.8g",pt.x,pt.y,pt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetDisplayCoordinates(ClientData,Tcl_Interp *interp,
		    int argc,CONST84 char** argv)
{ // Import: (x,y,z) in mesh coordinates
  // Export: (x,y,z) in display coordinates
  // The returned results are REAL4 precision.
  Tcl_ResetResult(interp);
  if(argc!=4) {
    Oc_Snprintf(buf,sizeof(buf),
		"GetMeshCellSize must be called with"
		" 3 arguments: x y z, in display coordinates"
		" (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL4 x = static_cast<REAL4>(Nb_Atof(argv[1]));
  REAL4 y = static_cast<REAL4>(Nb_Atof(argv[2]));
  REAL4 z = static_cast<REAL4>(Nb_Atof(argv[3]));
  Nb_Vec3<REAL4> pt(x,y,z);
  myFrame.CoordinatePointTransform(CS_CALCULATION_STANDARD,
				   myFrame.GetCoordinates(),
				   pt);
  Oc_Snprintf(buf,sizeof(buf),"%.8g %.8g %.8g",pt.x,pt.y,pt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int FindMeshVector(ClientData,Tcl_Interp *interp,
		    int argc,CONST84 char** argv)
{ // Import: x y z, in mesh coordinates
  // Export: x y z vx vy vz, in mesh coordinates,
  //   with viewaxis transformation applied.
  // The returned results are REAL8 precision.
  Tcl_ResetResult(interp);
  if(argc!=4) {
    Oc_Snprintf(buf,sizeof(buf),
		"FindMeshVector must be called with"
		" 3 arguments: x y z, in mesh coordinates"
		" (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL8m x = Nb_Atof(argv[1]);
  REAL8m y = Nb_Atof(argv[2]);
  REAL8m z = Nb_Atof(argv[3]);
  Nb_Vec3<REAL8> pos(x,y,z);
  Nb_LocatedVector<REAL8> lv;
  myMesh->FindPreciseClosest(pos,lv);
  lv.value *= myMesh->GetValueMultiplier();

  // viewaxis coordinate transform
  Oc_AutoBuf ab1,ab2;
  const char* viewaxis
    = Tcl_GetVar2(interp, ab1("plot_config"),
		  ab2("viewaxis"),TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
  if(viewaxis==NULL) return TCL_ERROR;
  if(strcmp("+z",viewaxis)!=0 && strcmp("z",viewaxis)!=0) {
    static CHAR cmdbuf[MY_BUF_SIZE];
    Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		"ApplyAxisTransform %s +z %.17g %.17g %.17g",
		viewaxis,lv.location.x,lv.location.y,lv.location.z);
    if(Tcl_Eval(interp,cmdbuf)!=TCL_OK) return TCL_ERROR;
    if(lv.location.Set(Tcl_GetStringResult(interp))!=0) {
      Oc_Snprintf(buf,sizeof(buf),
		  "Import string to lv.location.Set"
		  " not a numeric triplet: %s",
		  Tcl_GetStringResult(interp));
      Tcl_SetResult(interp,buf,TCL_VOLATILE);
      return TCL_ERROR;
    }
    Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		"ApplyAxisTransform %s +z %.17g %.17g %.17g",
		viewaxis,lv.value.x,lv.value.y,lv.value.z);
    if(Tcl_Eval(interp,cmdbuf)!=TCL_OK) return TCL_ERROR;
    if(lv.value.Set(Tcl_GetStringResult(interp))!=0) {
      Oc_Snprintf(buf,sizeof(buf),
		  "Import string to lv.value.Set"
		  " not a numeric triplet: %s",
		  Tcl_GetStringResult(interp));
      Tcl_SetResult(interp,buf,TCL_VOLATILE);
      return TCL_ERROR;
    }
  }

  // Save and return
  Oc_Snprintf(buf,sizeof(buf),"%.17g %.17g %.17g %.17g %.17g %.17g",
	      lv.location.x,lv.location.y,lv.location.z,
	      lv.value.x,lv.value.y,lv.value.z);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int GetMeshIncrement(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshIncrement must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),"%f",myMesh->GetSubsampleGrit());
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshSpatialUnitString(ClientData,Tcl_Interp *interp,
		  int argc,CONST84 char** argv)
{ // Returns string denoting mesh spatial units
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s\"", argv[0]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),"%s",myMesh->GetMeshUnit());
  if(strcmp("unknown",buf)==0) buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshRange(ClientData,Tcl_Interp *interp,
		  int argc,CONST84 char** /* argv */)
{ // Returns 6-tuple: xmin ymin zmin xmax ymax zmax, in mesh units.
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshRange must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_Vec3<REAL8> minpt,maxpt;
  Nb_BoundingBox<REAL8> range;
  myMesh->GetPreciseRange(range);
  range.GetExtremes(minpt,maxpt);
  Oc_Snprintf(buf,sizeof(buf),"%.17g %.17g %.17g %.17g %.17g %.17g",
	      minpt.x,minpt.y,minpt.z,maxpt.x,maxpt.y,maxpt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshZRange(ClientData,Tcl_Interp *interp,
		  int argc,CONST84 char** /* argv */)
{ // Returns a pair: zmin zmax.  These are in mesh units.
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetMeshZRange must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_Vec3<REAL8> minpt,maxpt;
  Nb_BoundingBox<REAL8> range;
  myMesh->GetPreciseRange(range);
  REAL8m zmin=0.,zmax=0.;
  if(!range.IsEmpty()) {
    range.GetExtremes(minpt,maxpt);
    zmin=minpt.z;
    zmax=maxpt.z;
  }
  Oc_Snprintf(buf,sizeof(buf),"%.17g %.17g",zmin,zmax);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetZsliceCount(ClientData,Tcl_Interp *interp,
		   int argc,CONST84 char** /* argv */)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetZsliceCount must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m slicecount = myMesh->GetZsliceCount();

  Oc_Snprintf(buf,sizeof(buf),"%d",slicecount);
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}

int GetZsliceLevels(ClientData,Tcl_Interp *interp,
		    int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=3) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetZsliceLevels must be called with 2 arguments:"
	    " z-low z-high (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL8m zlow =Nb_Atof(argv[1]);
  REAL8m zhigh =Nb_Atof(argv[2]);
  INT4m islicelow,islicehigh;
  myMesh->GetZslice(zlow,zhigh,islicelow,islicehigh);

  Oc_Snprintf(buf,sizeof(buf),"%d %d",islicelow,islicehigh);
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}


int ChangeMesh(ClientData cd,Tcl_Interp *interp,
	       int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<5 || argc>6) {
    Oc_Snprintf(buf,sizeof(buf),
            "ChangeMesh must be called with 4 or 5 arguments:"
	    " mesh_filename"
	    " frame_width frame_height rotation_degrees"
	    " [zoom]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  if(activeMeshId<0 || activeMeshId>=MY_MESH_ARRAY_SIZE) {
    Oc_Snprintf(buf,sizeof(buf),
	    "PROGRAMMING ERROR: activeMeshId=%d is out-of-range: [0,%d]",
	    activeMeshId,MY_MESH_ARRAY_SIZE-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  if(myMesh!=myMeshArray[activeMeshId]) {
    Oc_Snprintf(buf,sizeof(buf),
	    "PROGRAMMING ERROR: myMesh not in sync with myMeshArray.");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  delete myMeshArray[activeMeshId];
  myMeshArray[activeMeshId]=myMesh=NULL; // Safety
  if(argv[1]==NULL || argv[1][0]=='\0') { // Null request
    myMeshArray[activeMeshId]=new Vf_EmptyMesh();
  }
  else {
    Vf_FileInput *vffreader
      = Vf_FileInput::NewReader(argv[1]);
    if(vffreader==NULL) { // Unknown file type
      myMeshArray[activeMeshId]=new Vf_EmptyMesh();
    }
    else { // Reader found for file argv[1]
      myMeshArray[activeMeshId]=vffreader->NewMesh();
      delete vffreader;
      if(myMeshArray[activeMeshId]==NULL)
	myMeshArray[activeMeshId]=new Vf_EmptyMesh(); // Safety
    }
  }
  myMesh = myMeshArray[activeMeshId];

  myFrame.SetMesh(myMesh);

  REAL4m width=static_cast<REAL4m>(Nb_Atof(argv[2]));
  REAL4m height=static_cast<REAL4m>(Nb_Atof(argv[3]));
  CoordinateSystem coords =
    AngleToCoords(static_cast<REAL4m>(Nb_Atof(argv[4])));
  REAL4m zoom=-1.;
  if(argc>5) {
    zoom=static_cast<REAL4m>(Nb_Atof(argv[5]));
  }

  myFrame.SetCoordinates(coords);

  if(zoom>0.) SetZoom(zoom);
  else        SetZoom(interp,width,height);

  UpdatePlotConfiguration(cd,interp,0,NULL);
  /// For now, just use current configuration /**/

  return TCL_OK;
}

int DrawFrame(ClientData,Tcl_Interp *interp,
	      int argc,CONST84 char** argv)
{
  /*
   * We use the following buffer to store commands for passing to
   * Tcl_Eval().  It is important that we do not use the global
   * buffer, buf, for commands evaluted by Tcl_Eval() since that is
   * used to write result strings to Tcl_AppendResult().  If the
   * contents of a buffer are overwritten while Tcl_Eval() is
   * evaluating them, bad things can happen.
   *
   * Note that similar problems can occur if the contents of
   * cmdbuf are overwritten while Tcl_Eval() is evaluate it.  This
   * means it is a bad thing for DrawFrame() to appear on the call
   * stack twice.  Take care that none of the Tcl_Eval()'s below make
   * another call to DrawFrame().
   */
  static CHAR cmdbuf[MY_BUF_SIZE];

  int error_code;
  Tcl_ResetResult(interp);
  if(argc!=3) {
    Oc_Snprintf(buf,sizeof(buf),
            "DrawFrame must be called with 2 arguments: "
	    "canvas_name SliceCompatMode (%d arguments passed)",
	    argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char* canvas=argv[1];
  int zslicecompat=atoi(argv[2]);

  Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
	      "{%s} delete all ;# update idletasks",canvas);
  /// Uncomment the "update idletasks" command to get a white
  /// screen between frames
  error_code=Tcl_Eval(interp,cmdbuf); // Clear canvas
  if(error_code!=TCL_OK) return error_code;

  // Redraw canvas
  if(zslicecompat) error_code=myFrame.Render(canvas,0);
  else error_code=myFrame.Render(canvas,1);  // Hide arrows and pixels
  if(error_code!=TCL_OK) {
    Oc_Snprintf(buf,sizeof(buf),
		"\nError from myFrame.Render(%s,0)",canvas);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return error_code;
  }

  Oc_Snprintf(cmdbuf,sizeof(cmdbuf),"InitializeSliceDisplay");
  error_code=Tcl_Eval(interp,cmdbuf);
  if(error_code!=TCL_OK) {
    Oc_Snprintf(buf,sizeof(buf),"\nError from InitializeSliceDisplay");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return error_code;
  }

  Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
	      "global DrawFrameCount; incr DrawFrameCount");
  error_code=Tcl_Eval(interp,cmdbuf); // Increment DrawFrameCount variable
  if(error_code!=TCL_OK) return error_code;

/* Adjust subsample setting to actual value.
 * Note: Any tracings on plot_config triggered as a result
 *  of the Tcl_SetVar2 calls will send the name used in the
 *  call (here, e.g., "plot_config arrow_subsample") to the
 *  trace callback, and the call will appear to be from the
 *  proc that called this routine (DrawFrame).  If that routine
 *  has not made plot_config visible through a 'global plot_config'
 *  command, then a "upvar plot_config" will *fail* in the trace
 *  callback.  To protect against this, trace callbacks on plot_config
 *  should have the global name hardcoded into the trace call, and the
 *  callback itself should use the hardcoded value instead of the
 *  "name" argument.  For example:
 *
 *            trace variable plot_config(arrow_subsample) w \
 *                "MyCallback plot_config"
 *  and
 *        proc MyCallback { globalname name elt op } {
 *                upvar #0 $globalname x
 *                ...
 *        }
 *
 *  (The "elt" value can be either hardcoded into globalname, or
 *  processed separately.)
 */
  Oc_AutoBuf ab1,ab2;
  REAL4m arrow_sample,pixel_sample;
  REAL4m arrow_sample_request,pixel_sample_request;
  myFrame.GetRequestedSubsampleRates(arrow_sample_request,
				     pixel_sample_request);
  myFrame.GetActualSubsampleRates(arrow_sample,pixel_sample);
  if(arrow_sample_request<0.) {
    arrow_sample*=-1;
    if(arrow_sample > REAL4m(-0.01)) arrow_sample = REAL4m(-0.01);
  }
  Oc_Snprintf(buf,sizeof(buf),"%g",arrow_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
	      buf,TCL_GLOBAL_ONLY);
  if(pixel_sample_request<0.) {
    pixel_sample*=-1;
    if(pixel_sample > REAL4m(-0.01)) pixel_sample = REAL4m(-0.01);
  }
  Oc_Snprintf(buf,sizeof(buf),"%g",pixel_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
	      buf,TCL_GLOBAL_ONLY);

  return TCL_OK;
}

int GetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),"GetFrameRotation must be called with"
	    " 1 argument: name of Tk's Frame Rotation variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m rotang=int(OC_ROUND(CoordsToAngle(myFrame.GetCoordinates())));

  Oc_Snprintf(buf,sizeof(buf),"%d",rotang);
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

// Following routine added by SLW May-1998, modified by MJD June-1998.
// It copies the plot configuration values from the Tcl variable
// plot_config into the corresponding Frame plot_config structures.
// Also sets quantitylist value from Frame.
int UpdatePlotConfiguration(ClientData,Tcl_Interp *interp,
			    int,CONST84 char **)
{
  Tcl_ResetResult(interp);
  // Ignore argument list; just read plot_config() directly and update
  // all fields.
  const char *cptr; int colorcount;
  Oc_AutoBuf ab1,ab2;

  // Variables to handle "out-of-plane rotations", i.e., axis
  // transforms.
  Nb_DString flipstr("+x:+y:+z"),colorquantity;
  REAL8m phase=0.;
  BOOL invert=0;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("viewaxis"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr) && strcmp(cptr,"+x:+y:+z")!=0) {
    char axisbuf[8];
    Oc_Snprintf(axisbuf,sizeof(axisbuf),"+z,%s",cptr);
    cptr = Tcl_GetVar2(interp,ab1("view_transform"),axisbuf,
		       TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      flipstr = cptr;
    }
  }

  PlotConfiguration arrow_config,pixel_config;
  myFrame.GetPlotConfiguration(arrow_config,pixel_config);

  // Arrow plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,status"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("arrow,colorcount"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) colorcount=atoi(cptr);
  if(colorcount<0) colorcount=arrow_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("arrow,colormap"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.colormap.Setup(colorcount,cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,colorphase"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) phase=Nb_Atof(cptr);
  else                     phase=0.;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,colorreverse"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) invert=atoi(cptr);
  else                     invert=0;

  myMesh->ColorQuantityTransform(flipstr,colorquantity,phase,invert,
                                 arrow_config.colorquantity,
				 arrow_config.phase,arrow_config.invert);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,autosample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) arrow_config.autosample=atoi(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.subsample=static_cast<REAL4m>(Nb_Atof(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,size"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    arrow_config.size=static_cast<REAL4m>(Nb_Atof(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,antialias"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) arrow_config.antialias=atoi(cptr);

  // Pixel plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,status"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) pixel_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("pixel,colorcount"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) colorcount=atoi(cptr);
  if(colorcount<0) colorcount=pixel_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("pixel,colormap"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr))
    pixel_config.colormap.Setup(colorcount,cptr);

  pixel_config.stipple="";
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,opaque"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    int opaque=atoi(cptr);
    if(!opaque) pixel_config.stipple="gray25";
  }

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,colorphase"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) phase=Nb_Atof(cptr);
  else                     phase=0.;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,colorreverse"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) invert=atoi(cptr);
  else                     invert=0;

  myMesh->ColorQuantityTransform(flipstr,colorquantity,phase,invert,
                                 pixel_config.colorquantity,
				 pixel_config.phase,pixel_config.invert);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,autosample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) pixel_config.autosample=atoi(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    pixel_config.subsample=static_cast<REAL4m>(Nb_Atof(cptr));
  }

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,size"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    pixel_config.size=static_cast<REAL4m>(Nb_Atof(cptr));
  }

  // Misc.
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,drawboundary"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr) && !atoi(cptr)) {
    myFrame.SetDrawBoundary(0);
  } else {
    myFrame.SetDrawBoundary(1);
  }

  BOOL bw_error;
  REAL8m boundary_width;
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,boundarywidth"),TCL_GLOBAL_ONLY);
  boundary_width = Nb_Atof(cptr,bw_error);
  if(!bw_error) myFrame.SetBoundaryWidth(boundary_width);

  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,boundarycolor"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) myFrame.SetBoundaryColor(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,background"),
		   TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) myFrame.SetBackgroundColor(cptr);

  // Get and set quantity type list
  Nb_DString* sp;
  Nb_List<Nb_DString> cqtypes;
  int cqtypec=myFrame.GetColorQuantityTypes(cqtypes);
  if (cqtypec>0) {
    char **cqtypev=new char*[cqtypec];
    int i;
    for(i=0, sp=cqtypes.GetFirst() ;
	sp!=NULL && i<cqtypec ; i++,sp=cqtypes.GetNext()) {
      cqtypev[i]=new char[sp->Length()+1];
      strcpy(cqtypev[i],sp->GetStr());
    }
    int itop=i;
    char *cqlist=Tcl_Merge(cqtypec,cqtypev);
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("quantitylist"),
		cqlist,TCL_GLOBAL_ONLY);
    Tcl_Free(cqlist);
    for(i=0;i<itop;i++) delete[] cqtypev[i];
    delete[] cqtypev;
  } else {
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("quantitylist"),
		ab1(""),TCL_GLOBAL_ONLY);
  }

  // Try to make selected color quantities valid
  int acq_match=0,pcq_match=0;
  const char* dfcq=NULL;
  for(sp=cqtypes.GetFirst() ; sp!=NULL ; sp=cqtypes.GetNext()) {
    if(strcmp(arrow_config.colorquantity,sp->GetStr())==0) acq_match=1;
    if(strcmp(pixel_config.colorquantity,sp->GetStr())==0) pcq_match=1;
    if(dfcq==NULL) dfcq=sp->GetStr();
  }
  if(!acq_match && dfcq!=NULL) {
    arrow_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }
  if(!pcq_match && dfcq!=NULL) {
    pixel_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }

  // Apply new configuration
  myFrame.SetPlotConfiguration(arrow_config,pixel_config);
  return TCL_OK;
}

int SetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),"SetFrameRotation must be called with"
	    " 1 argument: new_rotation_angle"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  myFrame.SetCoordinates(AngleToCoords(
                         static_cast<REAL4m>(Nb_Atof(argv[1]))));

  return TCL_OK;
}

int GetFrameBox(ClientData,Tcl_Interp *interp,
		 int argc,CONST84 char**)
{ // NOTE: The return value from this routine is a *6*-tuple,
  //  "xmin ymin zmin xmax ymax zmax".
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetFrameBox should be called with no arguments."
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_BoundingBox<REAL4> bb=myFrame.GetDisplayBox();
  Nb_Vec3<REAL4> minpt,maxpt;
  bb.GetExtremes(minpt,maxpt);
  if(bb.IsEmpty()) maxpt=minpt;  // Empty boxes may have maxpt<minpt.

  Oc_Snprintf(buf,sizeof(buf),"%g %g %g %g %g %g",
	  minpt.x,minpt.y,minpt.z,maxpt.x,maxpt.y,maxpt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetZoom(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),"GetZoom must be called with"
	    " 1 argument: name of zoom variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Oc_Snprintf(buf,sizeof(buf),"%.6g",myFrame.GetZoom());
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

REAL4m SetZoom(REAL4m zoom)
{
  return myFrame.SetZoom(zoom);
}

REAL4m SetZoom(Tcl_Interp *interp,REAL4m width,REAL4m height)
{
  // Get margin and scroll information
  REAL4m margin=0.0,scrollbar_cross_dimension=0.0;
  Oc_AutoBuf ab1,ab2;
  const char *cptr;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("misc,margin"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    margin=static_cast<REAL4m>(Nb_Atof(cptr));
  }
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("misc,scrollcrossdim"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    scrollbar_cross_dimension=static_cast<REAL4m>(Nb_Atof(cptr));
  }
  return myFrame.SetZoom(width,height,
			 margin,scrollbar_cross_dimension);
}

int SetZoom(ClientData,Tcl_Interp *interp,
	    int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<2 || argc>3) {
    Oc_Snprintf(buf,sizeof(buf),"SetZoom must be called with"
	    " either 1 or 2 arguments: <new_zoom|new_width new_height>"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Change mesh scale in display list
  REAL4m newzoom;
  if(argc==2) newzoom=SetZoom(REAL4m(Nb_Atof(argv[1])));
  else        newzoom=SetZoom(interp,
			      REAL4m(Nb_Atof(argv[1])),
			      REAL4m(Nb_Atof(argv[2])));

  Oc_Snprintf(buf,sizeof(buf),"%g",newzoom);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int GetDefaultColorMapList(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),"GetDefaultColorMapList must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Nb_List<const char*> maps;
  DisplayColorMap::DefaultColorMapList(maps);
  INT4m mapcount = maps.GetSize();
  if(mapcount>0) {
    char **mapv=new char*[mapcount];
    int i;
    const char** cpp;
    Nb_List_Index<const char*> key;
    for(i=0, cpp=maps.GetFirst(key);
	i<mapcount && cpp!=NULL ;
	i++, cpp=maps.GetNext(key)) {
      mapv[i] = new char[strlen(*cpp)+1];
      strcpy(mapv[i],*cpp);
    }
    char* maplist=Tcl_Merge(mapcount,mapv);
    Tcl_AppendResult(interp,maplist,(char *)NULL);
    Tcl_Free(maplist);
    while(i>0) delete[] mapv[--i];
    delete[] mapv;
  }
  return TCL_OK;
}

int GetAutosamplingRates(ClientData,Tcl_Interp *interp,
			 int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),"GetAutosamplingRates must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  REAL4m arrow_rate
    = myFrame.GetAutoSampleRate(myFrame.GetPreferredArrowCellsize());
  REAL4m pixel_rate
    = myFrame.GetAutoSampleRate(myFrame.GetPreferredPixelCellsize());
  Oc_Snprintf(buf,sizeof(buf),"%.17g %.17g",arrow_rate,pixel_rate);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}


int GetDataValueUnit(ClientData,Tcl_Interp *interp,
		     int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),"GetDataValueUnit must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Oc_Snprintf(buf,sizeof(buf),"%.*s",MY_BUF_SIZE,myFrame.GetValueUnit());
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}

int GetDataValueScaling(ClientData,Tcl_Interp *interp,
			int argc,CONST84 char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Oc_Snprintf(buf,sizeof(buf),
            "GetDataValueScaling must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Oc_Snprintf(buf,sizeof(buf),"%.6g",myFrame.GetValueScaling());
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}


int SetDataValueScaling(ClientData,Tcl_Interp *interp,
		      int argc,CONST84 char** argv)
{
  // Adjust data value scaling
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),
            "SetDataValueScaling must be called with"
	    " 1 argument: <new_scale>"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL8m newscale=myFrame.SetValueScaling(Nb_Atof(argv[1]));
  Oc_Snprintf(buf,sizeof(buf),"%g",newscale);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int WriteMeshUsingDeprecatedVIOFormat(ClientData,Tcl_Interp *interp,
	      int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc != 2) {
    Oc_Snprintf(buf,sizeof(buf),
            "WriteMeshDeprecatedVIOFormat must be called with"
	    " 1 argument: filename"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  // Open file
  Tcl_Channel channel;
  Tcl_DString saveTranslation;
  if (argv[1][0] == '\0') {
    int mode;
    channel = Tcl_GetChannel(interp, Oc_AutoBuf("stdout"), &mode);
    Tcl_DStringInit(&saveTranslation);
    Tcl_GetChannelOption(interp, channel, Oc_AutoBuf("-translation"),
                         &saveTranslation);
  } else {
    channel = Tcl_OpenFileChannel(interp, argv[1], Oc_AutoBuf("w"), 0666);
  }
  if (channel == NULL) {
    return TCL_ERROR;
  }
  Tcl_SetChannelOption(interp, channel, Oc_AutoBuf("-translation"),
                       Oc_AutoBuf("lf"));

  Vf_VioFileOutput vfo;
  if(vfo.WriteMesh(myMesh,channel,NULL)!=0) {
      Oc_Snprintf(buf,sizeof(buf),"WriteMeshDeprecatedVIOFormat error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
  }
  if (argv[1][0] != '\0') {
    Tcl_Close(NULL, channel);
  } else {
    Tcl_Flush(channel);
    Tcl_SetChannelOption(interp, channel, Oc_AutoBuf("-translation"),
        Tcl_DStringValue(&saveTranslation));
    Tcl_DStringFree(&saveTranslation);
  }

  return errcode;
}

int WriteMesh(ClientData,Tcl_Interp *interp,
	      int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<4 || argc>6) {
    Oc_Snprintf(buf,sizeof(buf),"WriteMesh must be called with"
	    " 3-5 arguments: filename <text|binary4|binary8>"
	    " <regular|irregular> [title] [description]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  Vf_OvfDataStyle ods=vf_oascii;
  if(strcmp("binary4",argv[2])==0)      ods=vf_obin4;
  else if(strcmp("binary8",argv[2])==0) ods=vf_obin8;

  BOOL force_irreg=0;
  if(strcmp("irregular",argv[3])==0) force_irreg=1;

  const char *title="",*desc="";
  if(argc>4) title=argv[4];
  if(argc>5) desc=argv[5];

  // Open file
  Tcl_Channel channel;
  Tcl_DString saveTranslation;
  if (argv[1][0] == '\0') {
    int mode;
    channel = Tcl_GetChannel(interp, Oc_AutoBuf("stdout"), &mode);
    Tcl_DStringInit(&saveTranslation);
    Tcl_GetChannelOption(interp, channel, Oc_AutoBuf("-translation"),
                         &saveTranslation);
  } else {
    channel = Tcl_OpenFileChannel(interp, argv[1], Oc_AutoBuf("w"), 0666);
  }
  if (channel == NULL) {
    return TCL_ERROR;
  }
  Tcl_SetChannelOption(interp, channel, Oc_AutoBuf("-translation"),
                       Oc_AutoBuf("lf"));

  Vf_OvfFileOutput ofo;
  if(ofo.WriteMesh(myMesh,channel,ods,force_irreg,title,desc)!=0) {
      Oc_Snprintf(buf,sizeof(buf),"WriteMesh error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
  }
  if (argv[1][0] != '\0') {
    Tcl_Close(NULL, channel);
  } else {
    Tcl_Flush(channel);
    Tcl_SetChannelOption(interp, channel, Oc_AutoBuf("-translation"),
        Tcl_DStringValue(&saveTranslation));
    Tcl_DStringFree(&saveTranslation);
  }

  return errcode;
}

////////////////////////////////////////////////////////////////////////
// Support classes for WriteMeshAverages ///////////////////////////////

class TripleIndex {
public:
  // The Step*() member functions increment i, j, and k as
  // appropriate for a single forward step.  The return
  // value is 0 when the step stays inside the inner
  // loop range, 1 when stepping outside the inner
  // loop to the middle loop, 2 when stepping outside
  // the middle loop to the outer loop, and 3 when all
  // iterates are complete.
  //   The existing member functions are designed with Fortran
  // loop order, i.e., the natural (storage) order is i
  // increments fastest, then j, and finally k.  Different
  // members step through the values differently, however
  // --- this is the raison d'etre of the TripleIndex
  // class.  If one wants the usual C ordering, then either
  // introduce additional members, or just interchange i and k.
  //   The Reset*() member functions set i, j, and k to
  // just before the first in-range value.  This means that
  // the Step*() functions should be called before using
  // i, j, and k.
  //   The loop parameters i, j, k are initialized to be
  // just in front of the first in-range value.  This means
  // that Step* should be called before using the x, y, and
  // z values.  The Reset*() members can be used to reset
  // i, j, and k to their initialized values.
  //   Example usage for volume or point averaging:
  //
  //      foo.ResetXYZ();
  //      const int& i = foo.iref();
  //      const int& j = foo.jref();
  //      const int& k = foo.kref();
  //      while(StepXYZ()<3) {
  //           do_stuff(i,j,k);
  //           write_stuff // If point averaging
  //      }
  //      write_stuff // If volume averaging
  //
  // For x-line averaging:
  //
  //      foo.ResetXYZ();
  //      const int& i = foo.iref();
  //      const int& j = foo.jref();
  //      const int& k = foo.kref();
  //      int stepcontrol = foo.StepXYZ();
  //      while(stepcontrol<3) {
  //         do {
  //   	        do_stuff(i,j,k);
  //         } while((stepcontrol=foo.StepXYZ())==0);
  //         write_stuff;
  //      }
  //
  // For z-plane averaging:
  //
  //      foo.ResetXYZ();
  //      const int& i = foo.iref();
  //      const int& j = foo.jref();
  //      const int& k = foo.kref();
  //      int stepcontrol = foo.StepXYZ();
  //      while(stepcontrol<3) {
  //         do {
  //   	        do_stuff(i,j,k);
  //         } while((stepcontrol=foo.StepXYZ())<2);
  //         write_stuff;
  //      }
  //
  // In order to do ?-plane or ?-line averaging, one can
  // setup a member function pointer to the appropriate
  // Step???() member function.
  //
  TripleIndex(int set_imin,int set_imax,int set_isize,
	      int set_jmin,int set_jmax,int set_jsize,
	      int set_kmin,int set_kmax,int set_ksize);
  ~TripleIndex() {}

  // Preferred access order.  Use for volume and pointwise
  // averaging, also for x-lines or z-planes
  void ResetXYZ(); // Increments i first, then j, and last k
  int StepXYZ();
  int LineStepXYZ();
  int PlaneStepXYZ();

  // Use for y-lines
  void ResetYXZ(); // Increments j first, then i, and last k
  int StepYXZ();
  int LineStepYXZ();

  // Use for z-lines
  void ResetZXY(); // Increments k first, then i, and last j
  int StepZXY();
  int LineStepZXY();

  // Use for x-planes
  void ResetYZX(); // Increments j first, then k, and last i
  int StepYZX();
  int PlaneStepYZX();

  // Use for y-planes
  void ResetXZY(); // Increments i first, then k, and last j
  int StepXZY();
  int PlaneStepXZY();

  // The reference member functions return references to the
  // class index variables.  These values may change across
  // calls to Step and Reset.
  const int& iref() const { return i; }
  const int& jref() const { return j; }
  const int& kref() const { return k; }

private:
  int i,j,k;
  int imin,imax,isize;
  int jmin,jmax,jsize;
  int kmin,kmax,ksize;
};

TripleIndex::TripleIndex
(int set_imin,int set_imax,int set_isize,
 int set_jmin,int set_jmax,int set_jsize,
 int set_kmin,int set_kmax,int set_ksize) :
  i(-1),j(-1),k(-1),
  imin(set_imin),imax(set_imax),isize(set_isize),
  jmin(set_jmin),jmax(set_jmax),jsize(set_jsize),
  kmin(set_kmin),kmax(set_kmax),ksize(set_ksize)
{
  if(imin<0 || imin>imax || imax>isize) {
    OC_THROW("Error in TripleIndex initializer: Bad i value(s)");
  }
  if(jmin<0 || jmin>jmax || jmax>jsize) {
    OC_THROW("Error in TripleIndex initializer: Bad j value(s)");
  }
  if(kmin<0 || kmin>kmax || kmax>ksize) {
    OC_THROW("Error in TripleIndex initializer: Bad k value(s)");
  }
}
void TripleIndex::ResetXYZ()
{
  i=imin-1; j=jmin; k=kmin  ;
}

int TripleIndex::StepXYZ()
{
  if((++i)<imax) return 0;
  i=imin;  if((++j)<jmax) return 1;
  j=jmin;  if((++k)<kmax) return 2;
  return 3;
}

int TripleIndex::LineStepXYZ()
{
  if((++j)<jmax) return 1;
  j=jmin;  if((++k)<kmax) return 2;
  return 3;
}

int TripleIndex::PlaneStepXYZ()
{
  if((++k)<kmax) return 2;
  return 3;
}

void TripleIndex::ResetYXZ()
{
  i=imin; j=jmin-1; k=kmin;
}

int TripleIndex::StepYXZ()
{
  if((++j)<jmax) return 0;
  j=jmin;  if((++i)<imax) return 1;
  i=imin;  if((++k)<kmax) return 2;
  return 3;
}

int TripleIndex::LineStepYXZ()
{
  if((++i)<imax) return 1;
  i=imin;  if((++k)<kmax) return 2;
  return 3;
}

void TripleIndex::ResetZXY()
{
  i=imin; j=jmin; k=kmin-1;
}

int TripleIndex::StepZXY()
{
  if((++k)<kmax) return 0;
  k=kmin;  if((++i)<imax) return 1;
  i=imin;  if((++j)<jmax) return 2;
  return 3;
}

int TripleIndex::LineStepZXY()
{
  if((++i)<imax) return 1;
  i=imin;  if((++j)<jmax) return 2;
  return 3;
}

void TripleIndex::ResetYZX()
{
  i=imin; j=jmin-1; k=kmin;
}

int TripleIndex::StepYZX()
{
  if((++j)<jmax) return 0;
  j=jmin;  if((++k)<kmax) return 1;
  k=kmin;  if((++i)<imax) return 2;
  return 3;
}

int TripleIndex::PlaneStepYZX()
{
  if((++i)<imax) return 2;
  return 3;
}

void TripleIndex::ResetXZY()
{
  i=imin-1; j=jmin; k=kmin;
}

int TripleIndex::StepXZY()
{
  if((++i)<imax) return 0;
  i=imin;  if((++k)<kmax) return 1;
  k=kmin;  if((++j)<jmax) return 2;
  return 3;
}

int TripleIndex::PlaneStepXZY()
{
  if((++j)<jmax) return 2;
  return 3;
}

class LinearTripleIndex {
  // This class is conceptually similiar to the TripleIndex
  // class, except here the output is a single index, "offset"
  // into a linear representation of a 3D array.  The array
  // ordering is interpreted as FORTRAN ordering, i.e., the
  // natural access pattern is to increment i first, then j,
  // and last k.
  //  Example usage for volume or ptwise averaging:
  //
  //      foo.SetStepXYZ();
  //      const UINT4m& offset = foo.offsetref();
  //      while(Step()<3) {
  //           do_stuff(offset);
  //           write_stuff // If point averaging
  //      }
  //      write_stuff // If volume averaging
  //
  // For x-line averaging:
  //
  //      foo.SetStepXYZ();
  //      const UINT4m& offset = foo.offsetref();
  //      int stepcontrol = foo.Step();
  //      while(stepcontrol<3) {
  //         do {
  //   	        do_stuff(offset);
  //         } while((stepcontrol=foo.Step())==0);
  //         write_stuff;
  //      }
  //
  // For z-plane averaging:
  //
  //      foo.SetStepXYZ();
  //      const UINT4m& offset = foo.offsetref();
  //      int stepcontrol = foo.Step();
  //      while(stepcontrol<3) {
  //         do {
  //   	        do_stuff(offset);
  //         } while((stepcontrol=foo.Step())<2);
  //         write_stuff;
  //      }
  //
public:
  LinearTripleIndex(int set_imin,int set_imax,int set_isize,
		    int set_jmin,int set_jmax,int set_jsize,
		    int set_kmin,int set_kmax);
  ~LinearTripleIndex() {}
  void SetStepXYZ();
  void SetStepXZY();
  void SetStepYZX();
  void SetStepYXZ();
  void SetStepZXY();
  void Reset();
  UINT4m Step();
  void SaveIndex() { a_save=a; b_save=b; c_save=c; }
  void RestoreIndex();
  const UINT4m& offsetref() const { return offset; }
private:
  int imin,imax,isize;
  int jmin,jmax,jsize;
  int kmin,kmax;
  UINT4m offset;
  INT4m a,b,c;
  INT4m a_save,b_save,c_save;
  INT4m amin,amax,astep;
  INT4m bmin,bmax,bstep;
  INT4m cmin,cmax,cstep;
};

LinearTripleIndex::LinearTripleIndex
(int set_imin,int set_imax,int set_isize,
 int set_jmin,int set_jmax,int set_jsize,
 int set_kmin,int set_kmax) :
  imin(set_imin),imax(set_imax),isize(set_isize),
  jmin(set_jmin),jmax(set_jmax),jsize(set_jsize),
  kmin(set_kmin),kmax(set_kmax),
  offset(0),a(-1),b(-1),c(-1),a_save(-1),b_save(-1),c_save(-1),
  amin(-1),amax(-1),astep(-1),
  bmin(-1),bmax(-1),bstep(-1),
  cmin(-1),cmax(-1),cstep(-1)
{
  if(imin<0 || imin>imax || imax>isize) {
    OC_THROW("Error in LinearTripleIndex initializer: Bad i value(s)");
  }
  if(jmin<0 || jmin>jmax || jmax>jsize) {
    OC_THROW("Error in LinearTripleIndex initializer: Bad j value(s)");
  }
  if(kmin<0 || kmin>kmax) {
    OC_THROW("Error in LinearTripleIndex initializer: Bad k value(s)");
  }
}

void LinearTripleIndex::Reset()
{
  a = amin - 1;
  b = bmin;
  c = cmin;
  offset = a*astep + b*bstep + c*cstep;
  // offset is unsigned, so if amin=bmin=cmin=0, the
  // initial value for offset will be 0 - astep.  This
  // should be okay, as long as (0 - astep) + astep = 0,
  // since the latter should be the first value actually used.
}

void LinearTripleIndex::RestoreIndex()
{
  a=a_save; b=b_save; c=c_save;
  offset = a*astep + b*bstep + c*cstep;
}

void LinearTripleIndex::SetStepXYZ()
{
  amin=imin; amax=imax;  astep=1;
  bmin=jmin; bmax=jmax;  bstep=isize;
  cmin=kmin; cmax=kmax;  cstep=isize*jsize;
  Reset();
}

void LinearTripleIndex::SetStepXZY()
{
  amin=imin; amax=imax;  astep=1;
  bmin=kmin; bmax=kmax;  bstep=isize*jsize;
  cmin=jmin; cmax=jmax;  cstep=isize;
  Reset();
}

void LinearTripleIndex::SetStepYZX()
{
  amin=jmin; amax=jmax;  astep=isize;
  bmin=kmin; bmax=kmax;  bstep=isize*jsize;
  cmin=imin; cmax=imax;  cstep=1;
  Reset();
}

void LinearTripleIndex::SetStepYXZ()
{
  amin=jmin; amax=jmax;  astep=isize;
  bmin=imin; bmax=imax;  bstep=1;
  cmin=kmin; cmax=kmax;  cstep=isize*jsize;
  Reset();
}

void LinearTripleIndex::SetStepZXY()
{
  amin=kmin; amax=kmax;  astep=isize*jsize;
  bmin=imin; bmax=imax;  bstep=1;
  cmin=jmin; cmax=jmax;  cstep=isize;
  Reset();
}

UINT4m LinearTripleIndex::Step()
{
  if(++a<amax) {
    offset += astep;
    return 0;
  }
  a = amin;
  offset -= (amax-1-amin)*astep;
  if(++b<bmax) {
   offset += bstep;
   return 1;
  }
  b = bmin;
  offset -= (bmax-1-bmin)*bstep;
  if(++c<cmax) {
    offset += cstep;
    return 2;
  }
  return 3;
}


int
WriteMeshAverages(ClientData, Tcl_Interp *interp,
		  int argc, CONST84 char** argv)
{
  Nb_DString colhdrfmt,linefmt;
// #define DEFAULT_NUM_FMT "%- #24.17g"
# define DEFAULT_NUM_FMT "%- #20.15g"
  Tcl_Channel channel;
  int mode;
  Oc_AutoBuf ab1,ab2;
  const char* cptr;
  // Import config array:
  //   average  -- one of (space|plane|line|point|ball) (required)
  //   axis     -- one of (x|y|z)                       (required)
  //   ball_radius -- floating point value denoting radius
  //        of averaging ball, in problem coordinates.  Required
  //        iff average is "ball".
  //   range    -- 6-tuple xmin ymin zmin xmax ymax zmax (prob. coords.)
  //   rrange   -- 6-tuple, xmin ymin zmin xmax ymax zmax,
  //                each in range [0,1] (relative coordinates)
  //   normalize -- 1 or 0.  If 1, then each output point is
  //        divided by the maximum magnitude that would occur
  //        if all vectors in the manifold are aligned.  Thus
  //        the magnitude of the output vectors are all <=1.
  //   header   -- one of (fullhead|shorthead|nohead) (required)
  //   trailer  -- one of (tail|notail)               (required)
  //   numfmt   -- Format for numeric output default is DEFAULT_NUM_FMT
  //   descript -- description string
  //   index    -- 3-tuple: label units value
  //   vallab   -- Value label.  Default is "M".
  //   valfuncs -- List of triplets.  Each triplet is label unit
  //        expr-expression where "label" and "unit" are headers for a
  //        column in the output data table.  "expr-expression" is a
  //        Tcl expr expression which is applied point-by-point on the
  //        input file before any averaging is done.  Available
  //        variables are x, y, z, vx, vy, vz, vmag.  A couple of
  //        examples are
  //                    Ms A/m $vmag
  //                    M110 A/m {($Mx+$My)*0.70710678}
  //   defaultvals -- 1 or 0.  If 1, the vx, vy, and vz are included
  //        automatically in output table.  If 0, then only columns
  //        specified by the "functions" option are output.  In all
  //        cases the point coordinates are output.
  //
  // The active volume is set from range, if set.  Otherwise,
  // rrange is used.  If neither is set, the default is the
  // entire mesh volume.

  Tcl_ResetResult(interp);
  if(argc!=3) {
    Oc_Snprintf(buf,sizeof(buf),
            "WriteMeshAverages must be called 2 arguments:"
	    " output channel, config array name;"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Output channel
  if ((channel = Tcl_GetChannel(interp, argv[1], &mode)) == NULL) {
    return TCL_ERROR;
  }
  if (!(mode & TCL_WRITABLE)) {
    Oc_Snprintf(buf,sizeof(buf),"%s is not a writable channel",argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }


  // Averaging type
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("average"),TCL_LEAVE_ERR_MSG);
  if(cptr==NULL) return TCL_ERROR;
  enum { point, line, plane, space, ball } avetype = space;
  if(strcmp(cptr,"point")==0) {
    avetype = point;
  } else if(strcmp(cptr,"line")==0) {
    avetype = line;
  } else if(strcmp(cptr,"plane")==0) {
    avetype = plane;
  } else if(strcmp(cptr,"ball")==0) {
    avetype = ball;
  } else if(strcmp(cptr,"space")!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid average type string: %s;"
	    " Should be one of point, line, plane or space",
	    cptr);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Average type axis spec
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("axis"),TCL_LEAVE_ERR_MSG);
  if(cptr==NULL) return TCL_ERROR;
  if(cptr[1]!='\0' ||
     (cptr[0]!='x' && cptr[0]!='y' && cptr[0]!='z')) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid axis spec: %s;"
	    " Should be x, y, or z",
	    cptr);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  char axis = cptr[0];

    // Is this a rectangular mesh?
  if(strcmp(myMesh->GetMeshType(),"Vf_GridVec3f")!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid mesh type: %s\n",
		myMesh->GetMeshType());
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Downcast
  Vf_GridVec3f& grid = *((Vf_GridVec3f *)myMesh);
  REAL8m scale=grid.GetValueMultiplier();

  // Check dimensions
  int i,j,k,isize,jsize,ksize;
  grid.GetDimens(isize,jsize,ksize);
  if(isize<1 || jsize<1 || ksize<1) {
    Oc_Snprintf(buf,sizeof(buf),"Bad mesh dimensions: %d %d %d",
		isize,jsize,ksize);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  UINT4m whole_mesh_size = UINT4m(isize)*UINT4m(jsize)*UINT4m(ksize);

  Nb_Vec3<REAL8> base=grid.GetBasePoint();
  Nb_Vec3<REAL8> step=grid.GetGridStep();

  // Range select
  REAL8m xmin = base.x - step.x/2.;
  REAL8m ymin = base.y - step.y/2.;
  REAL8m zmin = base.z - step.z/2.;
  REAL8m xmax = base.x + (isize-0.5)*step.x;
  REAL8m ymax = base.y + (jsize-0.5)*step.y;
  REAL8m zmax = base.z + (ksize-0.5)*step.z;
  if(xmin>xmax) { REAL8m temp=xmin; xmin=xmax; xmax=temp; }
  if(ymin>ymax) { REAL8m temp=ymin; ymin=ymax; ymax=temp; }
  if(zmin>zmax) { REAL8m temp=zmin; zmin=zmax; zmax=temp; }

  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("range"),0);
  if(cptr!=NULL) {
    // range specified
    int range_arr_size;
    CONST84 char ** range_arr;
    if (Tcl_SplitList(interp,ab1(cptr),&range_arr_size,&range_arr)!=TCL_OK) {
      return TCL_ERROR;
    }
    if(range_arr_size!=6) {
      Tcl_Free((char *)range_arr);
      Oc_Snprintf(buf,sizeof(buf),"Range list has %d != 6 elements",
		  range_arr_size);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    if(strcmp("-",range_arr[0])!=0) {
      REAL8m temp = Nb_Atof(range_arr[0]);
      if(temp>xmin) xmin=temp;
    }
    if(strcmp("-",range_arr[1])!=0) {
      REAL8m temp = Nb_Atof(range_arr[1]);
      if(temp>ymin) ymin=temp;
    }
    if(strcmp("-",range_arr[2])!=0) {
      REAL8m temp = Nb_Atof(range_arr[2]);
      if(temp>zmin) zmin=temp;
    }
    if(strcmp("-",range_arr[3])!=0) {
      REAL8m temp = Nb_Atof(range_arr[3]);
      if(temp<xmax) xmax=temp;
    }
    if(strcmp("-",range_arr[4])!=0) {
      REAL8m temp = Nb_Atof(range_arr[4]);
      if(temp<ymax) ymax=temp;
    }
    if(strcmp("-",range_arr[5])!=0) {
      REAL8m temp = Nb_Atof(range_arr[5]);
      if(temp<zmax) zmax=temp;
    }
    Tcl_Free((char *)range_arr);
  } else if((cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("rrange"),0))!=NULL) {
    // rrange specified
    int range_arr_size;
    CONST84 char ** range_arr;
    if (Tcl_SplitList(interp,ab1(cptr),&range_arr_size,&range_arr)!=TCL_OK) {
      return TCL_ERROR;
    }
    if(range_arr_size!=6) {
      Tcl_Free((char *)range_arr);
      Oc_Snprintf(buf,sizeof(buf),"Rrange list has %d != 6 elements",
		  range_arr_size);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    REAL8m txmin=xmin,txmax=xmax;
    REAL8m tymin=ymin,tymax=ymax;
    REAL8m tzmin=zmin,tzmax=zmax;
    if(strcmp("-",range_arr[0])!=0) {
      REAL8m temp = Nb_Atof(range_arr[0]);
      if(temp>0) txmin = (1.0-temp)*xmin + temp*xmax;
    }
    if(strcmp("-",range_arr[1])!=0) {
      REAL8m temp = Nb_Atof(range_arr[1]);
      if(temp>0) tymin = (1.0-temp)*ymin + temp*ymax;
    }
    if(strcmp("-",range_arr[2])!=0) {
      REAL8m temp = Nb_Atof(range_arr[2]);
      if(temp>0) tzmin = (1.0-temp)*zmin + temp*zmax;
    }
    if(strcmp("-",range_arr[3])!=0) {
      REAL8m temp = Nb_Atof(range_arr[3]);
      if(temp<1) txmax = (1.0-temp)*xmin + temp*xmax;
    }
    if(strcmp("-",range_arr[4])!=0) {
      REAL8m temp = Nb_Atof(range_arr[4]);
      if(temp<1) tymax = (1.0-temp)*ymin + temp*ymax;
    }
    if(strcmp("-",range_arr[5])!=0) {
      REAL8m temp = Nb_Atof(range_arr[5]);
      if(temp<1) tzmax = (1.0-temp)*zmin + temp*zmax;
    }
    Tcl_Free((char *)range_arr);
    xmin = txmin;  xmax = txmax;
    ymin = tymin;  ymax = tymax;
    zmin = tzmin;  zmax = tzmax;
  }
  /// Otherwise, if neither range or rrange specified, default is full span

  // Scaling
  BOOL normalize = 0;
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("normalize"),0);
  if(cptr!=NULL) {
    if(atoi(cptr)) normalize=1;
  }

  // Default outputs?
  BOOL defaultvals = 1;
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("defaultvals"),0);
  if(cptr!=NULL) {
    if(!atoi(cptr)) defaultvals=0;
  }
  UINT4m value_column_count = ( defaultvals ? 3 : 0 );
  /// value_column_count is the number of output columns, excluding the
  /// index column (if any), and the x/y/z position columns (if any).

  // User supplied outputs
  Nb_ArrayWrapper<Nb_DString> user_labels,user_units,user_funcs;
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("valfuncs"),0);
  if(cptr!=NULL) {
    int valfuncs_count, errcode;
    CONST84 char **valfuncs_values;
    errcode = Tcl_SplitList(interp,cptr,&valfuncs_count,&valfuncs_values);
    if(errcode != TCL_OK) {
      return TCL_ERROR;
    }
    if(valfuncs_count%3!=0) {
      Tcl_Free((char *)valfuncs_values);
      Oc_Snprintf(buf,sizeof(buf),"valfuncs element count (%d)"
		  " not divisible by 3",valfuncs_count);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    user_labels.SetSize(valfuncs_count/3);
    user_units.SetSize(valfuncs_count/3);
    user_funcs.SetSize(valfuncs_count/3);
    for(i=0;i<valfuncs_count;i+=3) {
      // Use Nb_DString MergeArgs function to provide
      // embedded space protection in labels and units.
      Nb_DString dsarr[1];
      dsarr[0] = valfuncs_values[i];
      user_labels[i/3].MergeArgs(1,dsarr);
      dsarr[0] = valfuncs_values[i+1];
      user_units[i/3].MergeArgs(1,dsarr);
      user_funcs[i/3] = Nb_DString(valfuncs_values[i+2]);
    }
    Tcl_Free((char *)valfuncs_values);
  }
  value_column_count += user_funcs.GetSize();
  Nb_ArrayWrapper<REAL8m> value_column_results(value_column_count);

  // Index column
  Nb_DString index_label = "";
  Nb_DString index_units = "";
  Nb_DString index_value = "";
  const char* index_fmt = "%s";
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("index"),0);
  if(cptr!=NULL) {
    // index specified
    int arr_size;
    CONST84 char ** arr;
    if (Tcl_SplitList(interp,ab1(cptr),&arr_size,&arr)!=TCL_OK) {
      return TCL_ERROR;
    }
    if(arr_size!=3) {
      Tcl_Free((char *)arr);
      Oc_Snprintf(buf,sizeof(buf),"Index list has %d != 3 elements",
		  arr_size);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    // Copy info info index_* Nb_DString's.  Use Nb_DString MergeArgs
    // function to provide embedded space protection.
    Nb_DString dsarr[1];
    dsarr[0] = arr[0]; index_label.MergeArgs(1,dsarr);
    dsarr[0] = arr[1]; index_units.MergeArgs(1,dsarr);
    dsarr[0] = arr[2]; index_value.MergeArgs(1,dsarr);
    Tcl_Free((char *)arr);

    // Setup for spacing on fore and aft
    index_fmt = " %-10s ";
  }

  // Column "units" header info.  Use Nb_DString MergeArgs
  // function to provide embedded space protection.
  Nb_DString dstrarr[1];
  Nb_DString meshunit,valueunit;
  dstrarr[0] = grid.GetMeshUnit();
  meshunit.MergeArgs(1,dstrarr);
  if(normalize) dstrarr[0] = "";
  else          dstrarr[0] = grid.GetValueUnit();
  valueunit.MergeArgs(1,dstrarr);

  // Data table numeric format
  const char* numfmt = Tcl_GetVar2(interp,ab1(argv[2]),ab2("numfmt"),0);
  if(numfmt==NULL) numfmt = DEFAULT_NUM_FMT;

  // Column labels
  const char* vallab = Tcl_GetVar2(interp,ab1(argv[2]),ab2("vallab"),0);
  if(vallab==NULL) vallab = "M";
  Nb_DString xlab = vallab;   xlab.Append("_x");
  Nb_DString ylab = vallab;   ylab.Append("_y");
  Nb_DString zlab = vallab;   zlab.Append("_z");
  if(normalize || strcmp("",grid.GetValueUnit())==0) {
    xlab.ToLower();  ylab.ToLower();  zlab.ToLower();
  }

  // Header type
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("header"),TCL_LEAVE_ERR_MSG);
  if(cptr==NULL) return TCL_ERROR;
  enum { fullhead, shorthead, nohead } headtype = fullhead;
  if(strcmp(cptr,"shorthead")==0) {
    headtype = shorthead;
  } else if(strcmp(cptr,"nohead")==0) {
    headtype = nohead;
  } else if(strcmp(cptr,"fullhead")!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid header type string: %s;"
	    " Should be one of fullhead, shorthead or nohead",
	    cptr);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Trailer type
  cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("trailer"),TCL_LEAVE_ERR_MSG);
  if(cptr==NULL) return TCL_ERROR;
  enum { tail, notail } tailtype = tail;
  if(strcmp(cptr,"notail")==0) {
    tailtype = notail;
  } else if(strcmp(cptr,"tail")!=0) {
    Oc_Snprintf(buf,sizeof(buf),"Invalid trailer type string: %s;"
	    " Should be either tail or notail",
	    cptr);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Determine computation sub-grid
  if(xmin>xmax || ymin>ymax || zmin>zmax) {
    // Nothing to do.
    if(headtype != nohead) {
      Oc_Snprintf(buf,sizeof(buf),"## Empty range\n");
      Tcl_Write(channel, buf, int(strlen(buf)));
    }
    return TCL_OK;
    /// NOTE: Checking here on the ?min/?max float values
    ///  protects against some overflow conditions on the
    ///  parallel integer constructions below
  }
  int imin=0,imax=isize;
  if(step.x>0) {
    imin = int(ceil((xmin - base.x)/step.x));
    imax = int(floor((xmax - base.x)/step.x));
  } else if(step.x<0) {
    imin = int(ceil((xmax - base.x)/step.x));
    imax = int(floor((xmin - base.x)/step.x));
  }
  ++imax;
  if(imin<0)     imin = 0;
  if(imax>isize) imax = isize;
  int icount = imax - imin;

  int jmin=0,jmax=jsize;
  if(step.y>0) {
    jmin = int(ceil((ymin - base.y)/step.y));
    jmax = int(floor((ymax - base.y)/step.y));
  } else if(step.y<0) {
    jmin = int(ceil((ymax - base.y)/step.y));
    jmax = int(floor((ymin - base.y)/step.y));
  }
  ++jmax;
  if(jmin<0)     jmin = 0;
  if(jmax>jsize) jmax = jsize;
  int jcount = jmax - jmin;

  int kmin=0,kmax=ksize;
  if(step.z>0) {
    kmin = int(ceil((zmin - base.z)/step.z));
    kmax = int(floor((zmax - base.z)/step.z));
  } else if(step.z<0) {
    kmin = int(ceil((zmax - base.z)/step.z));
    kmax = int(floor((zmin - base.z)/step.z));
  }
  ++kmax;
  if(kmin<0)     kmin = 0;
  if(kmax>ksize) kmax = ksize;
  int kcount = kmax - kmin;

  if(icount<1 || jcount<1 || kcount<1) {
    // Nothing to do
    Oc_Snprintf(buf,sizeof(buf),"## Empty range\n");
    Tcl_Write(channel, buf, int(strlen(buf)));
    return TCL_OK;
  }
  unsigned int volume_point_count = icount*jcount*kcount;

  // Evaluate user supplied outputs.
  Nb_ArrayWrapper< Nb_ArrayWrapper<REAL8m> > uservals;
  uservals.SetSize(user_funcs.GetSize());
  UINT4m index;
  for(index=0;index<user_funcs.GetSize();index++) {
    uservals[index].SetSize(whole_mesh_size);
    UINT4m offset=kmin*jsize*isize;
    for(k=kmin;k<kmax;k++) {
      offset += jmin*isize;
      for(j=jmin;j<jmax;j++) {
	offset += imin;
	for(i=imin;i<imax;i++) {
	  Nb_Vec3<REAL8m>& v = grid(i,j,k);
	  Nb_Vec3<REAL8m> pos = grid.Position(i,j,k);
#if USE_TCLOBJS
#else // Use string interface
	  Oc_Snprintf(buf,sizeof(buf),
		      "set x %#.17g;  set y %#.17g;  set z %#.17g; "
		      "set vx %#.17g; set vy %#.17g; set vz %#.17g; "
		      "set vmag %#.17g; "
		      "expr {%s}",
		      pos.x,pos.y,pos.z,
		      scale*v.x,scale*v.y,scale*v.z,scale*v.Mag(),
		      user_funcs[index].GetStr());
	  // The '#' option in the format specifier is included to insure
	  // that a decimal point is output.  Otherwise Tcl may interpret
	  // some of these values as integers, which leads to oddities
	  // such as integer division and overflow.
	  if(Tcl_Eval(interp,buf) != TCL_OK) {
	    Tcl_AppendResult(interp,"\n---------------------"
			     "\nBad Tcl script:\n",buf,
			     "\n---------------------",(char *)NULL);
	    return TCL_ERROR;
	  }
	  BOOL bad_result;
	  uservals[index][offset]
	    = Nb_Atof(Tcl_GetStringResult(interp),bad_result);
	  if(bad_result) {
	    Tcl_AppendResult(interp,"\n---\nBad expr result: ",
			     Tcl_GetStringResult(interp),(char *)NULL);
	    return TCL_ERROR;
	  }
#endif
	  offset++;
	}
	offset += (isize-imax);
      }
      offset += (jsize-jmax)*isize;
    }
  }

  if(headtype != nohead) {
    // Print ODT file header
    Oc_Snprintf(buf,sizeof(buf),"# ODT 1.0\n");
    Tcl_Write(channel, buf, int(strlen(buf)));

    cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("descript"),0);
    if(cptr!=NULL) {
      Oc_Snprintf(buf,sizeof(buf),"## Desc: %s\n",cptr);
      Tcl_Write(channel, buf, int(strlen(buf)));
    }

    if(headtype == fullhead) {
      Oc_Snprintf(buf,sizeof(buf),
	"## Active volume: (%.15g,%.15g,%.15g) x (%.15g,%.15g,%.15g)\n"
	"## Cell size: %.15g x %.15g x %.15g\n"
	"## Cells in active volume: %u\n",
	xmin,ymin,zmin,xmax,ymax,zmax,
	fabs(step.x),fabs(step.y),fabs(step.z),
	volume_point_count);
      Tcl_Write(channel, buf, int(strlen(buf)));
    }

    Oc_Snprintf(buf,sizeof(buf),"#\n# Table Start\n");
    Tcl_Write(channel, buf, int(strlen(buf)));
  }

  char* bufptr=NULL;
  Nb_Vec3<REAL8m> m;
  if(avetype==space) {
    // Volume average
    if(headtype != nohead) {
      Oc_Snprintf(buf,sizeof(buf),
		  "# Title: Average across active volume\n");
      Tcl_Write(channel, buf, int(strlen(buf)));

      // Column headers
      bufptr = buf;
      bufptr += Oc_Snprintf(buf,sizeof(buf),"# Columns:\\\n# ");
      bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			    index_fmt,index_label.GetStr());
      if(index_label.Length()>0) *(bufptr++) = ' ';
      if(defaultvals) {
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s %-10s %-10s",xlab.GetStr(),
			      ylab.GetStr(),zlab.GetStr());
	if(user_labels.GetSize()>0) *(bufptr++) = ' ';
      }
      for(index=0;index<user_labels.GetSize();index++) {
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s",user_labels[index].GetStr());
	if(index+1<user_labels.GetSize()) *(bufptr++) = ' ';
      }
      *(bufptr++) = '\n';
      *bufptr = '\0'; // Safety
      Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));

      // Unit headers
      bufptr = buf;
      bufptr += Oc_Snprintf(buf,sizeof(buf),"#   Units:\\\n# ");
      bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			    index_fmt,index_units.GetStr());
      if(index_units.Length()>0) *(bufptr++) = ' ';
      if(defaultvals) {
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s %-10s %-10s",valueunit.GetStr(),
			      valueunit.GetStr(),valueunit.GetStr());
	if(user_units.GetSize()>0) *(bufptr++) = ' ';
      }
      for(index=0;index<user_units.GetSize();index++) {
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s",user_units[index].GetStr());
	if(index+1<user_units.GetSize()) *(bufptr++) = ' ';
      }
      *(bufptr++) = '\n';
      *bufptr = '\0'; // Safety
      Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
    }

    int user_col_offset = 0;
    if(defaultvals) {
      m.Set(0.,0.,0.);
      REAL8m magsum=0.0;
      for(k=kmin;k<kmax;k++) for(j=jmin;j<jmax;j++) for(i=imin;i<imax;i++) {
	m+=grid(i,j,k);
	if(normalize) magsum += grid(i,j,k).Mag();
      }
      if(normalize && magsum>0.0) m *= 1.0/magsum;
      else                        m *= scale/double(volume_point_count);
      value_column_results[0] = m.x;
      value_column_results[1] = m.y;
      value_column_results[2] = m.z;
      user_col_offset = 3;
    }
    for(index=0;index<user_funcs.GetSize();index++) {
      REAL8m sum=0.0;
      UINT4m kstep = jsize*isize;
      UINT4m koffset = kmin*kstep;
      for(k=kmin;k<kmax;k++,koffset+=kstep) {
	UINT4m joffset = koffset + jmin*isize;
	for(j=jmin;j<jmax;j++,joffset+=isize) {
	  for(i=imin;i<imax;i++) {
	    sum += uservals[index][joffset+i];
	  }
	}
      }
      value_column_results[user_col_offset+index]
	= sum/double(volume_point_count);
    }

    bufptr = buf;
    bufptr += Oc_Snprintf(buf,sizeof(buf)-1,
			  index_fmt,index_value.GetStr());
    if(bufptr!=buf) *(bufptr++) = ' ';
    for(index=0;index<value_column_count;index++) {
      bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			    numfmt,value_column_results[index]);
      if(index+1<value_column_count) {
	*(bufptr++) = ' ';
      }
    }
    *(bufptr++) = '\n';
    *bufptr = '\0'; // Safety
    Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
  } else {
    // Plane averages
    TripleIndex triple_index(imin,imax,isize,
			     jmin,jmax,jsize,
			     kmin,kmax,ksize);
    const int& iref = triple_index.iref();
    const int& jref = triple_index.jref();
    const int& kref = triple_index.kref();
    int (TripleIndex::* tripstep)() = NULL;

    REAL8m base_a=0,base_b=0,step_a=0,step_b=0;
    const int* a_index_ptr=NULL;
    const int* b_index_ptr=NULL;

    LinearTripleIndex linear_index(imin,imax,isize,
				   jmin,jmax,jsize,
				   kmin,kmax);
    const UINT4m& linear_offset = linear_index.offsetref();

    UINT4m manifold_size = 0; // Dummy values.  Actuals values
    REAL8m mult = -1.0;       // set inside switch case blocks.
    int stepcontrol = -1;     // (Compilers don't allow variable
    /// declarations at top-level scope.)
    switch(avetype) {
    case plane:
      switch(axis) {
	case 'x':
	  if(headtype != nohead) {
	    Oc_Snprintf(buf,sizeof(buf),
			"# Title: Averages across x-axis (%d points each)\n",
			jcount*kcount);
	  }
	  triple_index.ResetYZX();
	  if(defaultvals) tripstep = &TripleIndex::StepYZX;
	  else            tripstep = &TripleIndex::PlaneStepYZX;
	  linear_index.SetStepYZX();
	  manifold_size = jcount*kcount;
	  base_a = base.x;
	  step_a = step.x;
	  a_index_ptr = &iref;
	  break;
	case 'y':
	  if(headtype != nohead) {
	    Oc_Snprintf(buf,sizeof(buf),
			"# Title: Averages across y-axis (%d points each)\n",
			icount*kcount);
	  }
	  triple_index.ResetXZY();
	  if(defaultvals) tripstep = &TripleIndex::StepXZY;
	  else            tripstep = &TripleIndex::PlaneStepXZY;
	  linear_index.SetStepXZY();
	  manifold_size = icount*kcount;
	  base_a = base.y;
	  step_a = step.y;
	  a_index_ptr = &jref;
	  break;
	case 'z':
	  if(headtype != nohead) {
	    Oc_Snprintf(buf,sizeof(buf),
			"# Title: Averages across z-axis (%d points each)\n",
			icount*jcount);
	  }
	  triple_index.ResetXYZ();
	  if(defaultvals) tripstep = &TripleIndex::StepXYZ;
	  else            tripstep = &TripleIndex::PlaneStepXYZ;
	  linear_index.SetStepXYZ();
	  manifold_size = icount*jcount;
	  base_a = base.z;
	  step_a = step.z;
	  a_index_ptr = &kref;
	  break;
	default:
	  Oc_Snprintf(buf,sizeof(buf),"Invalid axis character: %c",axis);
	  Tcl_AppendResult(interp,buf,(char *)NULL);
	  return TCL_ERROR;
      }
      if(headtype != nohead) {
	Tcl_Write(channel, buf, int(strlen(buf))); // Title
	// Column headers
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf),"# Columns:\\\n# ");
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      index_fmt,index_label.GetStr());
	if(index_label.Length()>0) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10c ",axis);
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",xlab.GetStr(),
				ylab.GetStr(),zlab.GetStr());
	  if(user_labels.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_labels.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_labels[index].GetStr());
	  if(index+1<user_labels.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));

	// Unit headers
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf),"#   Units:\\\n# ");
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      index_fmt,index_units.GetStr());
	if(index_units.Length()>0) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s ",meshunit.GetStr());
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",valueunit.GetStr(),
				valueunit.GetStr(),valueunit.GetStr());
	  if(user_units.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_units.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_units[index].GetStr());
	  if(index+1<user_units.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
      }
      mult=(scale/double(manifold_size));
      if(defaultvals) stepcontrol = (triple_index.*tripstep)();
      else            stepcontrol = 0;
      linear_index.Step();
      while(stepcontrol<3) {
	int user_col_offset = 0;
	REAL8m position = base_a+double(*a_index_ptr)*step_a;
	/// Note: This is a bit of a hack.  In the !defaultvals case,
	/// the triple_index pointer is one back from the working
	/// location.  But that difference affects only the averaging
	/// variable, so doesn't influence this "position".
	if(defaultvals) {
	  m.Set(0.,0.,0.);
	  REAL8m magsum=0.0;
	  do {
	    m+=grid(iref,jref,kref);
	    if(normalize) magsum += grid(iref,jref,kref).Mag();
	  } while((stepcontrol=(triple_index.*tripstep)())<2);
	  if(normalize && magsum>0.0) m *= 1.0/magsum;
	  else                        m *= mult;
	  value_column_results[0] = m.x;
	  value_column_results[1] = m.y;
	  value_column_results[2] = m.z;
	  user_col_offset = 3;
	} else {
	  stepcontrol = (triple_index.*tripstep)();
	}
	linear_index.SaveIndex();
	for(index=0;index<user_funcs.GetSize();index++) {
	  linear_index.RestoreIndex();
	  REAL8m sum=0.0;
	  do {
	    sum += uservals[index][linear_offset];
	  } while(linear_index.Step()<2);
	  value_column_results[user_col_offset+index]
	    = sum/double(manifold_size);
	}
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf)-1,
			      index_fmt,index_value.GetStr()); // Index
	if(bufptr!=buf) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      numfmt,position); // x, y, or z
	*(bufptr++) = ' ';
	for(index=0;index<value_column_count;index++) { // Other values
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,value_column_results[index]);
	  if(index+1<value_column_count) {
	    *(bufptr++) = ' ';
	  }
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
      }
      break;
    case line:
      switch(axis) {
	case 'x':
	  if(headtype != nohead) {
	    bufptr = buf;
	    bufptr += Oc_Snprintf(buf,sizeof(buf),
				  "# Title: Averages parallel to"
				  " x-axis (%d points each)\n",icount);
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf),"# Columns:\\\n# ");
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  index_fmt,index_label.GetStr());
	    if(index_label.Length()>0) *(bufptr++) = ' ';
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  "%-10c %-10c ",'y','z');
	  }
	  triple_index.ResetXYZ();
	  if(defaultvals) tripstep = &TripleIndex::StepXYZ;
	  else            tripstep = &TripleIndex::LineStepXYZ;
	  linear_index.SetStepXYZ();
	  manifold_size = icount;
	  base_a = base.y;   step_a = step.y;   a_index_ptr = &jref;
	  base_b = base.z;   step_b = step.z;   b_index_ptr = &kref;
	  break;
	case 'y':
	  if(headtype != nohead) {
	    bufptr = buf;
	    bufptr += Oc_Snprintf(buf,sizeof(buf),
				  "# Title: Averages parallel to"
				  " y-axis (%d points each)\n",jcount);
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf),"# Columns:\\\n# ");
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  index_fmt,index_label.GetStr());
	    if(index_label.Length()>0) *(bufptr++) = ' ';
 	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  "%-10c %-10c ",'x','z');
	  }
	  triple_index.ResetYXZ();
	  if(defaultvals) tripstep = &TripleIndex::StepYXZ;
	  else            tripstep = &TripleIndex::LineStepYXZ;
	  linear_index.SetStepYXZ();
	  manifold_size = jcount;
	  base_a = base.x;   step_a = step.x;   a_index_ptr = &iref;
	  base_b = base.z;   step_b = step.z;   b_index_ptr = &kref;
	  break;
	case 'z':
	  if(headtype != nohead) {
	    bufptr = buf;
	    bufptr += Oc_Snprintf(buf,sizeof(buf),
				  "# Title: Averages parallel to"
				  " z-axis (%d points each)\n",kcount);
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf),"# Columns:\\\n# ");
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  index_fmt,index_label.GetStr());
	    if(index_label.Length()>0) *(bufptr++) = ' ';
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  "%-10c %-10c ",'x','y');
	  }
	  triple_index.ResetZXY();
	  if(defaultvals) tripstep = &TripleIndex::StepZXY;
	  else            tripstep = &TripleIndex::LineStepZXY;
	  linear_index.SetStepZXY();
	  manifold_size = kcount;
	  base_a = base.x;   step_a = step.x;   a_index_ptr = &iref;
	  base_b = base.y;   step_b = step.y;   b_index_ptr = &jref;
	  break;
	default:
	  Oc_Snprintf(buf,sizeof(buf),"Invalid axis character: %c",axis);
	  Tcl_AppendResult(interp,buf,(char *)NULL);
	  return TCL_ERROR;
      }
      if(headtype != nohead) {
	// Finish column headers
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",xlab.GetStr(),
				ylab.GetStr(),zlab.GetStr());
	  if(user_labels.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_labels.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_labels[index].GetStr());
	  if(index+1<user_labels.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	// Write title and column headers
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));

	// Unit headers
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf),"#   Units:\\\n# ");
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      index_fmt,index_units.GetStr());
	if(index_units.Length()>0) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s %-10s ",
			      meshunit.GetStr(),meshunit.GetStr());
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",valueunit.GetStr(),
				valueunit.GetStr(),valueunit.GetStr());
	  if(user_units.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_units.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_units[index].GetStr());
	  if(index+1<user_units.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	// Write units line to table
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
      }
      mult=(scale/double(manifold_size));
      if(defaultvals) stepcontrol = (triple_index.*tripstep)();
      else            stepcontrol = 0;
      linear_index.Step();
      while(stepcontrol<3) {
	int user_col_offset = 0;
	REAL8m position_a = base_a+double(*a_index_ptr)*step_a;
	REAL8m position_b = base_b+double(*b_index_ptr)*step_b;
	/// Note: This is a bit of a hack.  In the !defaultvals case,
	/// the triple_index pointer is one back from the working
	/// location.  But that difference affects only the averaging
	/// variable, so doesn't influence this "position".
	if(defaultvals) {
	  m.Set(0.,0.,0.);
	  REAL8m magsum=0.0;
	  do {
	    m+=grid(iref,jref,kref);
	    if(normalize) magsum += grid(iref,jref,kref).Mag();
	  } while((stepcontrol=(triple_index.*tripstep)())==0);
	  if(normalize && magsum>0.0) m *= 1.0/magsum;
	  else                        m *= mult;
	  value_column_results[0] = m.x;
	  value_column_results[1] = m.y;
	  value_column_results[2] = m.z;
	  user_col_offset = 3;
	} else {
	  stepcontrol = (triple_index.*tripstep)();
	}
	linear_index.SaveIndex();
	for(index=0;index<user_funcs.GetSize();index++) {
	  linear_index.RestoreIndex();
	  REAL8m sum=0.0;
	  do {
	    sum += uservals[index][linear_offset];
	  } while(linear_index.Step()==0);
	  value_column_results[user_col_offset+index]
	    = sum/double(manifold_size);
	}
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf)-1,
			      index_fmt,index_value.GetStr()); // Index
	if(bufptr!=buf) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      numfmt,position_a); // One of x, y, or z
	*(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      numfmt,position_b); // Another of y or z
	*(bufptr++) = ' ';
	for(index=0;index<value_column_count;index++) { // Other values
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,value_column_results[index]);
	  if(index+1<value_column_count) {
	    *(bufptr++) = ' ';
	  }
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
      }
      break;
    case point:
      // Point output (no averaging)
      if(headtype != nohead) {
	Oc_Snprintf(buf,sizeof(buf),
		    "# Title: Points in specified volume\n");
	Tcl_Write(channel, buf, int(strlen(buf)));

	// Column headers
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf),"# Columns:\\\n# ");
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      index_fmt,index_label.GetStr());
	if(index_label.Length()>0) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10c %-10c %-10c ",'x','y','z');
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",xlab.GetStr(),
				ylab.GetStr(),zlab.GetStr());
	  if(user_labels.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_labels.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_labels[index].GetStr());
	  if(index+1<user_labels.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));

	// Unit headers
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf),"#   Units:\\\n# ");
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      index_fmt,index_units.GetStr());
	if(index_units.Length()>0) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s %-10s %-10s ",meshunit.GetStr(),
			      meshunit.GetStr(),meshunit.GetStr());
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",valueunit.GetStr(),
				valueunit.GetStr(),valueunit.GetStr());
	  if(user_units.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_units.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_units[index].GetStr());
	  if(index+1<user_units.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
      }

      for(k=kmin;k<kmax;k++) {
	UINT4m joffset = (k*jsize + jmin)*isize;
	for(j=jmin;j<jmax;j++,joffset+=isize) for(i=imin;i<imax;i++) {
	  int user_col_offset = 0;
	  if(defaultvals) {
	    m=grid(i,j,k);
	    if(normalize) {
	      REAL8m mag = m.Mag();
	      if(mag>0.) { m.x /= mag;  m.y /= mag;  m.z /= mag; }
	    } else {
	      m *= scale;
	    }
	    value_column_results[0] = m.x;
	    value_column_results[1] = m.y;
	    value_column_results[2] = m.z;
	    user_col_offset = 3;
	  }
	  for(index=0;index<user_funcs.GetSize();index++) {
	    value_column_results[user_col_offset+index]
	      = uservals[index][joffset+i];
	  }

	  bufptr = buf;
	  bufptr += Oc_Snprintf(buf,sizeof(buf)-1,
				index_fmt,index_value.GetStr());
	  if(bufptr!=buf) *(bufptr++) = ' ';
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,base.x+double(i)*step.x);
	  *(bufptr++) = ' ';
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,base.y+double(j)*step.y);
	  *(bufptr++) = ' ';
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,base.z+double(k)*step.z);
	  *(bufptr++) = ' ';
	  for(index=0;index<value_column_count;index++) {
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  numfmt,value_column_results[index]);
	    if(index+1<value_column_count) {
	      *(bufptr++) = ' ';
	    }
	  }
	  *(bufptr++) = '\n';
	  *bufptr = '\0'; // Safety
	  Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
	}
      }
      break;
    case ball: {
      // Compute ball averaged values inside specified volume
      BOOL br_error;
      cptr=Tcl_GetVar2(interp,ab1(argv[2]),ab2("ball_radius"),0);
      REAL8m ball_radius = Nb_Atof(cptr,br_error);
      if(br_error) {
	Oc_Snprintf(buf,sizeof(buf),"Input error: ball averaging"
		    " radius not specified");
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
      REAL8m ball_radsq = ball_radius*ball_radius;
      int irad = int(floor(ball_radius/step.x));
      int jrad = int(floor(ball_radius/step.y));
      int krad = int(floor(ball_radius/step.z));

      // If input file is not 3D, then reduce ball dimension accordingly
      if(isize==1) irad=0;
      if(jsize==1) jrad=0;
      if(ksize==1) krad=0;

      // Count number of points (grid nodes) inside ball
      int ball_point_count = 0;
      for(k=-krad;k<=krad;k++) {
	REAL8m zoff = k*step.z;
	REAL8m zoffsq=zoff*zoff;
	for(j=-jrad;j<=jrad;j++) {
	  REAL8m yoff = j*step.y;
	  REAL8m yoffsq=yoff*yoff;
	  for(i=-irad;i<=irad;i++) {
	    REAL8m xoff = i*step.x;
	    REAL8m xoffsq=xoff*xoff;
	    if(zoffsq+yoffsq+xoffsq<=ball_radsq) {
	      ball_point_count++;
	    }
	  }
	}
      }

      if(headtype != nohead) {
	if(ball_point_count>1) {
	  Oc_Snprintf(buf,sizeof(buf),"# Title: Radius %g ball averages"
		      " across specified volume (%d points)\n",
		      ball_radius,ball_point_count);
	} else {
	  Oc_Snprintf(buf,sizeof(buf),"# Title: Radius %g ball averages"
		      " across specified volume (1 point)\n",
		      ball_radius);
	}
	Tcl_Write(channel, buf, int(strlen(buf)));

	// Column headers
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf),"# Columns:\\\n# ");
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      index_fmt,index_label.GetStr());
	if(index_label.Length()>0) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10c %-10c %-10c ",'x','y','z');
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",xlab.GetStr(),
				ylab.GetStr(),zlab.GetStr());
	  if(user_labels.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_labels.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_labels[index].GetStr());
	  if(index+1<user_labels.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));

	// Unit headers
	bufptr = buf;
	bufptr += Oc_Snprintf(buf,sizeof(buf),"#   Units:\\\n# ");
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      index_fmt,index_units.GetStr());
	if(index_units.Length()>0) *(bufptr++) = ' ';
	bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
			      "%-10s %-10s %-10s ",meshunit.GetStr(),
			      meshunit.GetStr(),meshunit.GetStr());
	if(defaultvals) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s %-10s %-10s",valueunit.GetStr(),
				valueunit.GetStr(),valueunit.GetStr());
	  if(user_units.GetSize()>0) *(bufptr++) = ' ';
	}
	for(index=0;index<user_units.GetSize();index++) {
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				"%-10s",user_units[index].GetStr());
	  if(index+1<user_units.GetSize()) *(bufptr++) = ' ';
	}
	*(bufptr++) = '\n';
	*bufptr = '\0'; // Safety
	Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
      }
      // Step through points about which a ball of radius ball_radius
      // fits entirely inside the active region.
      for(k=kmin+krad;k<kmax-krad;k++) for(j=jmin+jrad;j<jmax-jrad;j++)
	for(i=imin+irad;i<imax-irad;i++) {
	  int user_col_offset = 0;
	  if(defaultvals) {
	    m.Set(0.,0.,0.);
	    REAL8m magsum=0.0;
	    for(int k2=k-krad;k2<k+krad+1;k2++) {
	      REAL8m zoff = (k2-k)*step.z;
	      REAL8m zoffsq=zoff*zoff;
	      for(int j2=j-jrad;j2<j+jrad+1;j2++) {
		REAL8m yoff = (j2-j)*step.y;
		REAL8m yoffsq=yoff*yoff;
		for(int i2=i-irad;i2<i+irad+1;i2++) {
		  REAL8m xoff = (i2-i)*step.x;
		  REAL8m xoffsq=xoff*xoff;
		  if(zoffsq+yoffsq+xoffsq>ball_radsq) continue;
		  m+=grid(i2,j2,k2);
		  if(normalize) magsum += grid(i2,j2,k2).Mag();
		}
	      }
	    }
	    if(normalize && magsum>0.0) m *= 1.0/magsum;
	    else                        m *= scale/double(ball_point_count);
	    value_column_results[0] = m.x;
	    value_column_results[1] = m.y;
	    value_column_results[2] = m.z;
	    user_col_offset = 3;
	  }
	  for(index=0;index<user_funcs.GetSize();index++) {
	    REAL8m sum=0.0;
	    UINT4m kstep = jsize*isize;
	    UINT4m koffset = (k-krad)*kstep;
	    for(int k2=k-krad;k2<k+krad+1;k2++,koffset+=kstep) {
	      REAL8m zoff = (k2-k)*step.z;
	      REAL8m zoffsq=zoff*zoff;
	      UINT4m joffset = koffset + (j-jrad)*isize;
	      for(int j2=j-jrad;j2<j+jrad+1;j2++,joffset+=isize) {
		REAL8m yoff = (j2-j)*step.y;
		REAL8m yoffsq=yoff*yoff;
		for(int i2=i-irad;i2<i+irad+1;i2++) {
		  REAL8m xoff = (i2-i)*step.x;
		  REAL8m xoffsq=xoff*xoff;
		  if(zoffsq+yoffsq+xoffsq>ball_radsq) continue;
		  sum += uservals[index][joffset+i2];
		}
	      }
	    }
	    value_column_results[user_col_offset+index]
	      = sum/double(ball_point_count);
	  }

	  bufptr = buf;
	  bufptr += Oc_Snprintf(buf,sizeof(buf)-1,
				index_fmt,index_value.GetStr());
	  if(bufptr!=buf) *(bufptr++) = ' ';
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,base.x+double(i)*step.x);
	  *(bufptr++) = ' ';
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,base.y+double(j)*step.y);
	  *(bufptr++) = ' ';
	  bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				numfmt,base.z+double(k)*step.z);
	  *(bufptr++) = ' ';
	  for(index=0;index<value_column_count;index++) {
	    bufptr += Oc_Snprintf(bufptr,sizeof(buf)-(bufptr-buf)-1,
				  numfmt,value_column_results[index]);
	    if(index+1<value_column_count) {
	      *(bufptr++) = ' ';
	    }
	  }
	  *(bufptr++) = '\n';
	  *bufptr = '\0'; // Safety
	  Tcl_Write(channel, buf, static_cast<int>(bufptr-buf));
	}
      }
      break;
    default:
      Oc_Snprintf(buf,sizeof(buf),"Programming error: avetype %d"
		  " invalid or not properly handled",
		  int(avetype));
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
      // Implicit "break"
    }
  }

  if(tailtype != notail) {
    // Print ODT file trailer
    Oc_Snprintf(buf,sizeof(buf),"# Table End\n");
    Tcl_Write(channel, buf, int(strlen(buf)));
  }

  Tcl_Flush(channel);
  return TCL_OK;
#undef DEFAULT_NUM_FMT
}

int
BitmapCmd(ClientData clientData, Tcl_Interp *interp,
	  int argc, CONST84 char** argv)
{
  char c;
  const char* cptr;
  int margin, width, height, crop, isNew, mode;
  int xmin, ymin, xmax, ymax, xmarginadj, ymarginadj;
  int draw_boundary;
  REAL8m boundary_width;
  REAL8m mat_width=0.0;  // Mat width, in pixels
  Nb_DString mat_color = "0xFFFFFF";  // Default color is white
  OommfBitmap *bitmapPtr;
  Oc_AutoBuf ab1, ab2;
  Tcl_HashTable *tablePtr = (Tcl_HashTable *)clientData;
  Tcl_HashEntry *entryPtr;
  Tcl_Channel channel;

  CHAR cmdbuf[MY_BUF_SIZE];

  if (argc < 2) {
    Oc_Snprintf(buf,sizeof(buf),
		"wrong # args: should be \"%s option bitmapName"
		" ?arg ...?\"", argv[0]);
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }
  c = argv[1][0];
  if ((c == 'C') && (strcmp(argv[1], "Create") == 0)) {
    if (argc != 3) {
      Oc_Snprintf(buf,sizeof(buf),
		  "wrong # args: should be \"%s Create bitmapName\"",
		  argv[0]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    entryPtr = Tcl_CreateHashEntry(tablePtr, argv[2], &isNew);
    if (!isNew) {
      Oc_Snprintf(buf,sizeof(buf),
		  "bitmap \"%s\" already exists", argv[2]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }

    // Create a new OommfBitmap from the global DisplayFrame
    bitmapPtr = new OommfBitmap;

    margin = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			      ab2("misc,margin"), TCL_GLOBAL_ONLY));
    width = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			     ab2("misc,width"), TCL_GLOBAL_ONLY));
    height = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			      ab2("misc,height"), TCL_GLOBAL_ONLY));
    crop = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			    ab2("misc,crop"), TCL_GLOBAL_ONLY));

    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,drawboundary"),TCL_GLOBAL_ONLY);
    draw_boundary=1;
    if(!Nb_StrIsSpace(cptr)) draw_boundary = atoi(cptr);
    if(draw_boundary) myFrame.SetDrawBoundary(1);
    else              myFrame.SetDrawBoundary(0);

    BOOL bw_error;
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,boundarywidth"),TCL_GLOBAL_ONLY);
    boundary_width = Nb_Atof(cptr,bw_error);
    if(!bw_error) myFrame.SetBoundaryWidth(boundary_width);

    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,boundarycolor"),TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) myFrame.SetBoundaryColor(cptr);

    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,boundarypos"),TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      if(Nb_StrCaseCmp(cptr,"back")==0) {
	myFrame.SetBoundaryOnTop(0);
      } else if(Nb_StrCaseCmp(cptr,"front")==0) {
	myFrame.SetBoundaryOnTop(1);
      } else {
	Oc_Snprintf(buf,sizeof(buf),
		    "Invalid plot_config(misc,boundarypos) value:"
		    " \"%s\"; should be either \"front\" or \"back\"",
		    cptr);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
    }

    BOOL mw_error;
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,matwidth"),TCL_GLOBAL_ONLY);
    REAL8m mw_temp = Nb_Atof(cptr,mw_error);
    if(!mw_error) mat_width = mw_temp; // Default is 0.0
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		       ab2("misc,matcolor"),TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      mat_color = cptr;   // Default is OxFFFFFF (white)
    }

    // Mesh box
    Nb_Vec3<REAL8> minpt, maxpt;
    Nb_BoundingBox<REAL8> meshbox;
    myMesh->GetPreciseRange(meshbox);
    meshbox.GetExtremes(minpt,maxpt);

    REAL8 default_span = maxpt.z-minpt.z;
    if(IsRegularMesh(myMesh)) {
      Nb_Vec3<REAL4> celldim=myMesh->GetApproximateCellDimensions();
      default_span = celldim.z;
      if(default_span<=0.0) default_span=1.0; // Safety
      if(default_span>maxpt.z-minpt.z) {
	default_span = maxpt.z-minpt.z;
      } else if(20*default_span<maxpt.z-minpt.z) {
	default_span = (maxpt.z-minpt.z)/20.;
      }
    }

    // Set up slice selection
    char viewaxis = 'z';
    char viewdir[3] = "+z";
    BOOL negative_viewdir = 0;
    BOOL arrowspan_error,pixelspan_error;
    REAL8m arrowspan,pixelspan;
    cptr = Tcl_GetVar2(interp, ab1("plot_config"),
		       ab2("viewaxis"), TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      size_t strsize = strlen(cptr);
      if(strsize==1) {
	viewaxis = cptr[0];
	viewdir[0] = '+'; viewdir[1]=viewaxis; viewdir[2]='\0';
      } else if(strsize==2) {
	viewaxis = cptr[1];
	strcpy(viewdir,cptr);
	if(cptr[0]=='-') {
	  negative_viewdir = 1; // Viewing from the backside
	}
      } else {
	Oc_Snprintf(buf,sizeof(buf),
		    "Invalid plot_config(viewaxis) string : \"%s\"",
		    cptr);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
    }
    if(viewaxis!='x' && viewaxis!='y' && viewaxis!='z') {
      Oc_Snprintf(buf,sizeof(buf),
		  "Invalid viewaxis detected: %c; Check setting of"
		  " plot_config(viewaxis)",viewaxis);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }

    BOOL centerpt_set = 0;
    Nb_Vec3<REAL8m> centerpt;
    cptr = Tcl_GetVar2(interp, ab1("plot_config"),
		       ab2("misc,centerpt"), TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      if(centerpt.Set(cptr)!=0) {
	Oc_Snprintf(buf,sizeof(buf),
		    "Error processing plot_config(misc,centerpt): %s",
		    cptr);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
      centerpt_set = 1;
    } else {
      cptr = Tcl_GetVar2(interp, ab1("plot_config"),
			 ab2("misc,relcenterpt"), TCL_GLOBAL_ONLY);
      if(!Nb_StrIsSpace(cptr)) {
	if(centerpt.Set(cptr)!=0) {
	  Oc_Snprintf(buf,sizeof(buf),
		      "Error processing plot_config(misc,relcenterpt): %s",
		      cptr);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return TCL_ERROR;
	}
	// Convert mesh range from viewaxis to standard coords
	Nb_Vec3<REAL8> scmin,scmax;
	Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		    "ApplyAxisTransform %s +z %.17g %.17g %.17g",
		    viewdir,minpt.x,minpt.y,minpt.z);
	Tcl_SavedResult saved;
	Tcl_SaveResult(interp, &saved);
	int error_code=Tcl_Eval(interp,cmdbuf);
	if(error_code!=TCL_OK) {
	  Tcl_DiscardResult(&saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error processing minpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return error_code;
	}
	if(scmin.Set(Tcl_GetStringResult(interp))!=0) {
	  Tcl_RestoreResult(interp, &saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error reading minpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return TCL_ERROR;
	}
	Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		    "ApplyAxisTransform %s +z %.17g %.17g %.17g",
		    viewdir,maxpt.x,maxpt.y,maxpt.z);
	error_code=Tcl_Eval(interp,cmdbuf);
	if(error_code!=TCL_OK) {
	  Tcl_DiscardResult(&saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error processing maxpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return error_code;
	}
	if(scmax.Set(Tcl_GetStringResult(interp))!=0) {
	  Tcl_RestoreResult(interp, &saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error reading maxpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return TCL_ERROR;
	}
	Tcl_RestoreResult(interp, &saved);

	// Convert from relative to absolute problem coordinates
	if(negative_viewdir) {
	  Nb_Vec3<REAL8m> tmpvec;
	  tmpvec = scmin;  scmin = scmax;  scmax = tmpvec;
	}
	centerpt.x = (1-centerpt.x)*scmin.x + centerpt.x*scmax.x;
	centerpt.y = (1-centerpt.y)*scmin.y + centerpt.y*scmax.y;
	centerpt.z = (1-centerpt.z)*scmin.z + centerpt.z*scmax.z;
	centerpt_set = 1;
      }
    }

    if(centerpt_set) {
      // Convert from problem to viewaxis coords
      Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		  "ApplyAxisTransform +z %s %.17g %.17g %.17g",
		  viewdir,centerpt.x,centerpt.y,centerpt.z);
      Tcl_SavedResult saved;
      Tcl_SaveResult(interp, &saved);
      int error_code=Tcl_Eval(interp,cmdbuf);
      if(error_code!=TCL_OK) {
	Tcl_DiscardResult(&saved);
	Oc_Snprintf(buf,sizeof(buf),"Error processing centerpt: %s",
		    cmdbuf);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return error_code;
      }
      if(centerpt.Set(Tcl_GetStringResult(interp))!=0) {
	Tcl_RestoreResult(interp, &saved);
	Oc_Snprintf(buf,sizeof(buf),"Error reading centerpt: %s",
		    cmdbuf);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
      Tcl_RestoreResult(interp, &saved);
    }

    Oc_Snprintf(buf,sizeof(buf),"viewaxis,%carrowspan",viewaxis);
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),buf,TCL_GLOBAL_ONLY);
    arrowspan = Nb_Atof(cptr,arrowspan_error);
    if(arrowspan_error || arrowspan==0) {
      arrowspan = default_span;
    } else if(arrowspan<0.0) {
      arrowspan = (maxpt.z-minpt.z);
    }

    Oc_Snprintf(buf,sizeof(buf),"viewaxis,%cpixelspan",viewaxis);
    cptr = Tcl_GetVar2(interp,ab1("plot_config"),buf,TCL_GLOBAL_ONLY);
    pixelspan = Nb_Atof(cptr,pixelspan_error);
    if(pixelspan_error || pixelspan==0) {
      pixelspan = default_span;
    } else if(arrowspan<0.0) {
      pixelspan = (maxpt.z-minpt.z);
    }

    // Calculate bounding boxes, in mesh coords
    Nb_Vec3<REAL4> temp_min,temp_max;
    Convert(minpt,temp_min);  Convert(maxpt,temp_max);
    temp_min.z = static_cast<REAL4>(centerpt.z - arrowspan/2);
    temp_max.z = static_cast<REAL4>(centerpt.z + arrowspan/2);
    // Make shave adjustment to better mimic mmDisp output.
    if(negative_viewdir) temp_max.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
    else                 temp_min.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
    Nb_BoundingBox<REAL4> arrow_box(temp_min,temp_max);
    temp_min.z = static_cast<REAL4>(centerpt.z - pixelspan/2);
    temp_max.z = static_cast<REAL4>(centerpt.z + pixelspan/2);
    // Make shave adjustment to better mimic mmDisp output.
    if(negative_viewdir) temp_max.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
    else                 temp_min.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
    Nb_BoundingBox<REAL4> pixel_box(temp_min,temp_max);

    // Setup bitmap
    Nb_Vec3<REAL4> disp_minpt,disp_maxpt;
    Nb_BoundingBox<REAL4> bbox;
    bbox = myFrame.GetDisplayBox();
    bbox.GetExtremes(disp_minpt, disp_maxpt);

    // If centerpt is specified, then pan bitmap window
    REAL4 xadj=0.0,yadj=0.0;
    if(centerpt_set) {
      Nb_Vec3<REAL4> view_center,display_center;
      Convert(centerpt,view_center);
      display_center.x = static_cast<REAL4>((minpt.x+maxpt.x)/2.);
      display_center.y = static_cast<REAL4>((minpt.y+maxpt.y)/2.);
      display_center.z = static_cast<REAL4>((minpt.z+maxpt.z)/2.);
      myFrame.CoordinatePointTransform(CS_CALCULATION_STANDARD,
				       myFrame.GetCoordinates(),
				       view_center);
      myFrame.CoordinatePointTransform(CS_CALCULATION_STANDARD,
				       myFrame.GetCoordinates(),
				       display_center);
      xadj = view_center.x-display_center.x;  // Pan amount
      yadj = view_center.y-display_center.y;
    }
    xmin = int(floor(disp_minpt.x+xadj)) - margin;
    ymin = int(floor(disp_minpt.y+yadj)) - margin;
    xmax = int(ceil(disp_maxpt.x+xadj))  + margin;
    ymax = int(ceil(disp_maxpt.y+yadj))  + margin;

    // Check size
    xmarginadj = (xmax - xmin + 1) - width;
    ymarginadj = (ymax - ymin + 1) - height;
    if(xmarginadj>0) {
      // Resulting size is bigger than requested. Trim.
      xmin += int(floor(xmarginadj/2.));
      xmax -= int(ceil(xmarginadj/2.));
    }
    if(ymarginadj>0) {
      // Resulting size is bigger than requested. Trim.
      ymin += int(floor(ymarginadj/2.));
      ymax -= int(ceil(ymarginadj/2.));
    }
    if(!crop) {
      // No-crop request, so make bitmap exact size requested
      xmax = xmin + width - 1;
      ymax = ymin + height - 1;
    }

    bitmapPtr->Setup(xmin,ymin,xmax,ymax);

    // Fill bitmap
    myFrame.FillBitmap(arrow_box,pixel_box,negative_viewdir,*bitmapPtr);

    // Add mat, if requested
    if(mat_width>0) {
      OommfPackedRGB color(mat_color.GetStr());
      bitmapPtr->AddMat(mat_width,color);
    }

    Tcl_SetHashValue(entryPtr, (ClientData)bitmapPtr);
    return TCL_OK;
  }
  if ((c == 'D') && (strcmp(argv[1], "Delete") == 0)) {
    if (argc != 3) {
      Oc_Snprintf(buf,sizeof(buf),
		  "wrong # args: should be \"%s Delete bitmapName\"",
		  argv[0]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    entryPtr = Tcl_FindHashEntry(tablePtr, argv[2]);
    if (entryPtr == NULL) {
      Oc_Snprintf(buf,sizeof(buf), "can't find bitmap \"%s\"",
		  argv[2]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    delete (OommfBitmap *) Tcl_GetHashValue(entryPtr);
    Tcl_DeleteHashEntry(entryPtr);
    return TCL_OK;
  }
  if ((c == 'W') && (strcmp(argv[1], "Write") == 0)) {
    if (argc != 5) {
      Oc_Snprintf(buf,sizeof(buf), "wrong # args: should be "
		  "\"%s Write bitmapName format channel\"", argv[0]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    if ((channel = Tcl_GetChannel(interp, argv[4], &mode)) == NULL) {
      return TCL_ERROR;
    }
    if (!(mode & TCL_WRITABLE)) {
      Oc_Snprintf(buf,sizeof(buf),
		  "%s is not a writable channel", argv[4]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    entryPtr = Tcl_FindHashEntry(tablePtr, argv[2]);
    if (entryPtr == NULL) {
      Oc_Snprintf(buf,sizeof(buf),
		  "can't find bitmap \"%s\"", argv[2]);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    bitmapPtr = (OommfBitmap *) Tcl_GetHashValue(entryPtr);
    if ( strcmp("P3", argv[3]) == 0) {
      bitmapPtr->WritePPMChannel(channel, 3);
      return TCL_OK;
    }
    if ( strcmp("P6", argv[3]) == 0) {
      bitmapPtr->WritePPMChannel(channel, 6);
      return TCL_OK;
    }
    if ( strcmp("B24", argv[3]) == 0) {
      bitmapPtr->WriteBMPChannel(channel, 24);
      return TCL_OK;
    }
    Oc_Snprintf(buf,sizeof(buf),
		"ERROR: Unknown file format request: %s\n",argv[3]);
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }
  Oc_Snprintf(buf,sizeof(buf),
	      "bad option \"%s\": must be Create, Delete, or Write",
	      argv[1]);
  Tcl_AppendResult(interp, buf, (char *) NULL);
  return TCL_ERROR;
}

void
BitmapDeleteProc(ClientData clientData)
{
  Tcl_HashTable *tablePtr = (Tcl_HashTable *) clientData;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchId;
  OommfBitmap *bitmapPtr;
  for (entryPtr = Tcl_FirstHashEntry(tablePtr, &searchId);
       entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&searchId)) {
    bitmapPtr = (OommfBitmap *) Tcl_GetHashValue(entryPtr);
    delete bitmapPtr;
  }
  Tcl_DeleteHashTable(tablePtr);
  delete tablePtr;
}

int PSWriteMesh(ClientData,Tcl_Interp *interp,
		int argc,CONST84 char** argv)
{
  const char* cptr;

  Tcl_ResetResult(interp);
  if(argc!=2) {
    Oc_Snprintf(buf,sizeof(buf),"PSWriteMesh must be called with"
	    " 1 argument: channel"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int mode;
  Tcl_Channel channel;
  if ((channel = Tcl_GetChannel(interp, argv[1], &mode)) == NULL) {
    return TCL_ERROR;
  }
  if (!(mode & TCL_WRITABLE)) {
    Oc_Snprintf(buf,sizeof(buf),
		"%s is not a writable channel", argv[1]);
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }

  Oc_AutoBuf ab1,ab2;
  CHAR cmdbuf[MY_BUF_SIZE];

  int margin = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
				ab2("misc,margin"), TCL_GLOBAL_ONLY));
  int width = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			       ab2("misc,width"), TCL_GLOBAL_ONLY));
  if(width<1) width = 1;
  int height = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
				ab2("misc,height"), TCL_GLOBAL_ONLY));
  if(height<1) height = 1;
  int croptoview = atoi(Tcl_GetVar2(interp, ab1("print_config"),
				    ab2("croptoview"), TCL_GLOBAL_ONLY));
  int crop = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
			      ab2("misc,crop"), TCL_GLOBAL_ONLY));
  // If plot_config(misc,crop) is true, then the print area (bounding box)
  // is cropped to fit the display margins.  This control is active when
  // the zoom value is small.
  //   Conversely, if print_config(croptoview) is true then the
  // specified pan (center offset) values are ignored, and the display
  // window is sized to wrap the mesh at the specified zoom, which is
  // then scaled to the print area accordingly.  "croptoview" is a
  // legacy value from the interactive print dialog, where it is used
  // primarily for overriding big zoom values.

  cptr = Tcl_GetVar2(interp,ab1("print_config"),
		     ab2("pwidth"),TCL_GLOBAL_ONLY);
  REAL8m pwidth = Nb_Atof(cptr);
  cptr = Tcl_GetVar2(interp,ab1("print_config"),
		     ab2("pheight"),TCL_GLOBAL_ONLY);
  REAL8m pheight = Nb_Atof(cptr);
  cptr = Tcl_GetVar2(interp,ab1("print_config"),
		     ab2("tmargin"),TCL_GLOBAL_ONLY);
  REAL8m ptmargin = Nb_Atof(cptr);
  cptr = Tcl_GetVar2(interp,ab1("print_config"),
		     ab2("lmargin"),TCL_GLOBAL_ONLY);
  REAL8m plmargin = Nb_Atof(cptr);
  cptr = Tcl_GetVar2(interp,ab1("print_config"),
		     ab2("units"),TCL_GLOBAL_ONLY);
  REAL8m printscale = 72.; // Default is inches
  if(Nb_StrCaseCmp(cptr,"cm")==0) {
    printscale = 72./2.54;
  } else if(Nb_StrCaseCmp(cptr,"pt")==0) {
    printscale = 1.0;
  }
  pwidth *= printscale;
  pheight *= printscale;
  ptmargin *= printscale;
  plmargin *= printscale;

  // Determine page dimensions, in points (where 72 points = 1 inch)
  Nb_DString paper_type = Tcl_GetVar2(interp,ab1("print_config"),
				      ab2("paper"),TCL_GLOBAL_ONLY);
  paper_type.Trim();  paper_type.ToLower();
  REAL8m page_width  = 0.0, page_height = 0.0;
  if(strcmp("letter",paper_type.GetStr())==0) {
    page_width  = 72 *  8.5;
    page_height = 72 * 11;
  } else if(strcmp("legal",paper_type.GetStr())==0) {
    page_width  = 72 *  8.5;
    page_height = 72 * 14;
  } else if(strcmp("executive",paper_type.GetStr())==0) {
    page_width  = 72 *  7.5;
    page_height = 72 * 10;
  } else if(strcmp("tabloid",paper_type.GetStr())==0) {
    page_width  = 72 * 11;
    page_height = 72 * 17;
  } else if(strcmp("ledger",paper_type.GetStr())==0) {
    page_width  = 72 * 17;
    page_height = 72 * 11;
  } else if(strcmp("statement",paper_type.GetStr())==0) {
    page_width  = 72 * 5.5;
    page_height = 72 * 8.5;
  } else if(strcmp("folio",paper_type.GetStr())==0) {
    page_width  = 72 * 8.5;
    page_height = 72 * 13;
  } else if(strcmp("10x14",paper_type.GetStr())==0
	    || strcmp("10 x 14",paper_type.GetStr())==0) {
    page_width  = 72 * 10;
    page_height = 72 * 14;
  } else if(strcmp("quarto",paper_type.GetStr())==0) {
    page_width  = 610;
    page_height = 780;
  } else if(paper_type[0]=='a') {
    BOOL error;
    REAL8m n = Nb_Atof(paper_type.GetStr()+1,error);
    if(!error) {
      // The official A<N> sizes are fudged a bit from the simple
      // exponential formula to insure: (1) each paper dimension is
      // a whole number of millimeters, and (2) the width of each
      // paper size is not more than 1/2 the length of the next
      // larger size.  The "floor" and +0.22 fudge factor adjust
      // for these irregularities for N in the range [-2,10].
      page_width  = floor(1000.*(pow(2.,-0.5*n-0.25))+0.22)*72./25.4;
      page_height = floor(1000.*(pow(2.,-0.5*n+0.25))+0.22)*72./25.4;
    }
  } else if(paper_type[0]=='b') {
    BOOL error;
    REAL8m n = Nb_Atof(paper_type.GetStr()+1,error);
    if(!error) {
      // Note 1: Comments above for A<N> sizes apply to B<N> sizes too.
      // Note 2: Ghostview appears to have wrong values for B4 and B5.
      //  The correct values for the ISO B<N> series, in points, are
      //  709x1001 and 499x709, respectively, where 1 inch == 72 points
      //  (aka a "PostScript point", as opposed to a "Printer's point",
      //  tpt, where 72.27 tpt = 1 inch).  Ghostview uses 729x1032 and
      //  516x729, respectively.  Further poking around hints that
      //  ghostview has both "B<n>" and "ISO B<n>", the latter agreeing
      //  with the values quoted above.  But I haven't been to find any
      //  other source listing the Ghostview B<n> values.
      page_width  = floor(1000.*pow(2.,-0.5*n))*72./25.4;
      page_height = floor(1000.*pow(2.,-0.5*n+0.5))*72./25.4;
    }
  }
  page_width  = OC_ROUND(page_width);
  page_height = OC_ROUND(page_height);
  if(page_width<=0.0 || page_height<=0.0) {
    Oc_Snprintf(buf,sizeof(buf),
		"Unrecognized paper request: %s",paper_type.GetStr());
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }

  // Check orientation request
  Nb_DString page_orientation =
    Tcl_GetVar2(interp,ab1("print_config"),
		ab2("orient"),TCL_GLOBAL_ONLY);
  page_orientation.Trim(); page_orientation.ToLower();
  if(strcmp("portrait",page_orientation.GetStr())==0) {
    page_orientation = "Portrait";
  } else if(strcmp("landscape",page_orientation.GetStr())==0) {
    page_orientation = "Landscape";
  } else {
    Oc_Snprintf(buf,sizeof(buf),
		"Unrecognized orientation request: %s",
		page_orientation.GetStr());
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }

  // Compute print page offsets, in points
  Nb_DString page_hpos =
    Tcl_GetVar2(interp,ab1("print_config"),
		ab2("hpos"),TCL_GLOBAL_ONLY);
  page_hpos.Trim(); page_hpos.ToLower();
  Nb_DString page_vpos =
    Tcl_GetVar2(interp,ab1("print_config"),
		ab2("vpos"),TCL_GLOBAL_ONLY);
  page_vpos.Trim(); page_vpos.ToLower();

  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,drawboundary"),TCL_GLOBAL_ONLY);
  int draw_boundary=1;
  if(!Nb_StrIsSpace(cptr)) draw_boundary = atoi(cptr);
  if(draw_boundary) myFrame.SetDrawBoundary(1);
  else              myFrame.SetDrawBoundary(0);

  BOOL bw_error;
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,boundarywidth"),TCL_GLOBAL_ONLY);
  REAL8m boundary_width = Nb_Atof(cptr,bw_error);
  if(!bw_error) myFrame.SetBoundaryWidth(boundary_width);

  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,boundarycolor"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) myFrame.SetBoundaryColor(cptr);

  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,boundarypos"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    if(Nb_StrCaseCmp(cptr,"back")==0) {
      myFrame.SetBoundaryOnTop(0);
    } else if(Nb_StrCaseCmp(cptr,"front")==0) {
      myFrame.SetBoundaryOnTop(1);
    } else {
      Oc_Snprintf(buf,sizeof(buf),
		  "Invalid plot_config(misc,boundarypos) value:"
		  " \"%s\"; should be either \"front\" or \"back\"",
		  cptr);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
  }

  // Mat info
  BOOL mw_error;
  REAL8m mat_width = 0.0;
  Nb_DString mat_color = "0xFFFFFF";  // Default color is white
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,matwidth"),TCL_GLOBAL_ONLY);
  REAL8m mw_temp = Nb_Atof(cptr,mw_error);
  if(!mw_error) mat_width = mw_temp;
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("misc,matcolor"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    mat_color = cptr;   // Default is 0xFFFFFF (white)
  }

  // Arrow boundary (edge) parameters
  BOOL ab_error;
  REAL8m arrow_boundary_width = 0.0;  // Default is no boundary
  Nb_DString arrow_boundary_color = "0x000000";
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("arrow,boundarywidth"),TCL_GLOBAL_ONLY);
  REAL8m ab_temp = Nb_Atof(cptr,ab_error);
  if(!ab_error) arrow_boundary_width = ab_temp;
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),
		     ab2("arrow,boundarycolor"),TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    arrow_boundary_color = cptr;   // Default is 0x000000 (black)
  }

  // Mesh box
  Nb_Vec3<REAL8> minpt, maxpt;
  Nb_BoundingBox<REAL8> meshbox;
  myMesh->GetPreciseRange(meshbox);
  meshbox.GetExtremes(minpt,maxpt);

  REAL8 default_span = maxpt.z-minpt.z;
  if(IsRegularMesh(myMesh)) {
    Nb_Vec3<REAL4> celldim=myMesh->GetApproximateCellDimensions();
    default_span = celldim.z;
    if(default_span<=0.0) default_span=1.0; // Safety
    if(default_span>maxpt.z-minpt.z) {
      default_span = maxpt.z-minpt.z;
    } else if(20*default_span<maxpt.z-minpt.z) {
      default_span = (maxpt.z-minpt.z)/20.;
    }
  }

  // Set up slice selection
  char viewaxis = 'z';
  char viewdir[3] = "+z";
  BOOL negative_viewdir = 0;
  BOOL arrowspan_error,pixelspan_error;
  REAL8m arrowspan,pixelspan;
  cptr = Tcl_GetVar2(interp, ab1("plot_config"),
		     ab2("viewaxis"), TCL_GLOBAL_ONLY);
  if(!Nb_StrIsSpace(cptr)) {
    size_t strsize = strlen(cptr);
    if(strsize==1) {
      viewaxis = cptr[0];
      viewdir[0] = '+'; viewdir[1]=viewaxis; viewdir[2]='\0';
    } else if(strsize==2) {
      viewaxis = cptr[1];
      strcpy(viewdir,cptr);
      if(cptr[0]=='-') {
	negative_viewdir = 1; // Viewing from the backside
      }
    } else {
      Oc_Snprintf(buf,sizeof(buf),
		  "Invalid plot_config(viewaxis) string : \"%s\"",
		  cptr);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
  }
  if(viewaxis!='x' && viewaxis!='y' && viewaxis!='z') {
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid viewaxis detected: %c; Check setting of"
		" plot_config(viewaxis)",viewaxis);
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
  }

  BOOL centerpt_set = 0;
  Nb_Vec3<REAL8m> centerpt;
  if(croptoview) {
    cptr = Tcl_GetVar2(interp, ab1("plot_config"),
		       ab2("misc,centerpt"), TCL_GLOBAL_ONLY);
    if(!Nb_StrIsSpace(cptr)) {
      if(centerpt.Set(cptr)!=0) {
	Oc_Snprintf(buf,sizeof(buf),
		    "Error processing plot_config(misc,centerpt): %s",
		    cptr);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
      }
      centerpt_set = 1;
    } else {
      cptr = Tcl_GetVar2(interp, ab1("plot_config"),
			 ab2("misc,relcenterpt"), TCL_GLOBAL_ONLY);
      if(!Nb_StrIsSpace(cptr)) {
	if(centerpt.Set(cptr)!=0) {
	  Oc_Snprintf(buf,sizeof(buf),
		      "Error processing plot_config(misc,relcenterpt): %s",
		      cptr);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return TCL_ERROR;
	}
	// Convert mesh range from viewaxis to standard coords
	Nb_Vec3<REAL8> scmin,scmax;
	Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		    "ApplyAxisTransform %s +z %.17g %.17g %.17g",
		    viewdir,minpt.x,minpt.y,minpt.z);
	Tcl_SavedResult saved;
	Tcl_SaveResult(interp, &saved);
	int error_code=Tcl_Eval(interp,cmdbuf);
	if(error_code!=TCL_OK) {
	  Tcl_DiscardResult(&saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error processing minpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return error_code;
	}
	if(scmin.Set(Tcl_GetStringResult(interp))!=0) {
	  Tcl_RestoreResult(interp, &saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error reading minpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return TCL_ERROR;
	}
	Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		    "ApplyAxisTransform %s +z %.17g %.17g %.17g",
		    viewdir,maxpt.x,maxpt.y,maxpt.z);
	error_code=Tcl_Eval(interp,cmdbuf);
	if(error_code!=TCL_OK) {
	  Tcl_DiscardResult(&saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error processing maxpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return error_code;
	}
	if(scmax.Set(Tcl_GetStringResult(interp))!=0) {
	  Tcl_RestoreResult(interp, &saved);
	  Oc_Snprintf(buf,sizeof(buf),"Error reading maxpt: %s",
		      cmdbuf);
	  Tcl_AppendResult(interp, buf, (char *) NULL);
	  return TCL_ERROR;
	}
	Tcl_RestoreResult(interp, &saved);

	// Convert from relative to absolute problem coordinates
	if(negative_viewdir) {
	  Nb_Vec3<REAL8m> tmpvec;
	  tmpvec = scmin;  scmin = scmax;  scmax = tmpvec;
	}
	centerpt.x = (1-centerpt.x)*scmin.x + centerpt.x*scmax.x;
	centerpt.y = (1-centerpt.y)*scmin.y + centerpt.y*scmax.y;
	centerpt.z = (1-centerpt.z)*scmin.z + centerpt.z*scmax.z;
	centerpt_set = 1;
      }
    }
  }

  if(centerpt_set) {
    // Convert from problem to viewaxis coords
    Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
		"ApplyAxisTransform +z %s %.17g %.17g %.17g",
		viewdir,centerpt.x,centerpt.y,centerpt.z);
    Tcl_SavedResult saved;
    Tcl_SaveResult(interp, &saved);
    int error_code=Tcl_Eval(interp,cmdbuf);
    if(error_code!=TCL_OK) {
      Tcl_DiscardResult(&saved);
      Oc_Snprintf(buf,sizeof(buf),"Error processing centerpt: %s",
		  cmdbuf);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return error_code;
    }
    if(centerpt.Set(Tcl_GetStringResult(interp))!=0) {
      Tcl_RestoreResult(interp, &saved);
      Oc_Snprintf(buf,sizeof(buf),"Error reading centerpt: %s",
		  cmdbuf);
      Tcl_AppendResult(interp, buf, (char *) NULL);
      return TCL_ERROR;
    }
    Tcl_RestoreResult(interp, &saved);
  }

  Oc_Snprintf(buf,sizeof(buf),"viewaxis,%carrowspan",viewaxis);
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),buf,TCL_GLOBAL_ONLY);
  arrowspan = Nb_Atof(cptr,arrowspan_error);
  if(arrowspan_error || arrowspan==0) {
    arrowspan = default_span;
  } else if(arrowspan<0.0) {
    arrowspan = (maxpt.z-minpt.z);
  }

  Oc_Snprintf(buf,sizeof(buf),"viewaxis,%cpixelspan",viewaxis);
  cptr = Tcl_GetVar2(interp,ab1("plot_config"),buf,TCL_GLOBAL_ONLY);
  pixelspan = Nb_Atof(cptr,pixelspan_error);
  if(pixelspan_error || pixelspan==0) {
    pixelspan = default_span;
  } else if(arrowspan<0.0) {
    pixelspan = (maxpt.z-minpt.z);
  }

  // Setup image
  Nb_Vec3<REAL4> disp_minpt,disp_maxpt;
  Nb_BoundingBox<REAL4> bbox;
  bbox = myFrame.GetDisplayBox();
  bbox.GetExtremes(disp_minpt, disp_maxpt);

  // If centerpt is specified, then pan display window
  REAL4 xadj=0.0,yadj=0.0;
  if(centerpt_set) {
    Nb_Vec3<REAL4> view_center,display_center;
    Convert(centerpt,view_center);
    display_center.x = static_cast<REAL4>((minpt.x+maxpt.x)/2.);
    display_center.y = static_cast<REAL4>((minpt.y+maxpt.y)/2.);
    display_center.z = static_cast<REAL4>((minpt.z+maxpt.z)/2.);
    myFrame.CoordinatePointTransform(CS_CALCULATION_STANDARD,
				     myFrame.GetCoordinates(),
				     view_center);
    myFrame.CoordinatePointTransform(CS_CALCULATION_STANDARD,
				     myFrame.GetCoordinates(),
				     display_center);
    xadj = view_center.x-display_center.x;  // Pan amount
    yadj = view_center.y-display_center.y;
  }
  REAL8m xmin = disp_minpt.x + xadj - margin;
  REAL8m ymin = disp_minpt.y + yadj - margin;
  REAL8m xmax = disp_maxpt.x + xadj + margin;
  REAL8m ymax = disp_maxpt.y + yadj + margin;

  // Check size
  if(croptoview) {
    // Adjust data size to match print size
    REAL8m xmarginadj = (xmax - xmin + 1 - width)/2.;
    REAL8m ymarginadj = (ymax - ymin + 1 - height)/2.;
    xmin += xmarginadj;
    xmax -= xmarginadj;
    ymin += ymarginadj;
    ymax -= ymarginadj;
    if(crop) {
      // Remove any slop
      if(xmin<disp_minpt.x-margin) xmin=disp_minpt.x-margin;
      if(ymin<disp_minpt.y-margin) ymin=disp_minpt.y-margin;
      if(xmax>disp_maxpt.x+margin) xmax=disp_maxpt.x+margin;
      if(ymax>disp_maxpt.y+margin) ymax=disp_maxpt.y+margin;
      REAL8m newwidth  = xmax - xmin + 1;
      REAL8m newheight = ymax - ymin + 1;
      REAL8m newpwidth  = pwidth * newwidth  / width;
      REAL8m newpheight = pheight * newheight / height;
      width  = static_cast<int>(OC_ROUND(newwidth));
      height = static_cast<int>(OC_ROUND(newheight));
      pwidth  = newpwidth;
      pheight = newpheight;
    }
  }

  // Calculate bounding boxes, in mesh coords
  REAL8m arrowsize = myFrame.GetArrowSize();
  Nb_Vec3<REAL4> temp_min(static_cast<REAL4>(xmin-arrowsize/2),
			  static_cast<REAL4>(ymin-arrowsize/2),0.0f);
  Nb_Vec3<REAL4> temp_max(static_cast<REAL4>(xmax+arrowsize/2),
			  static_cast<REAL4>(ymax+arrowsize/2),0.0f);
  myFrame.CoordinatePointTransform(myFrame.GetCoordinates(),
				   CS_CALCULATION_STANDARD,
				   temp_min);
  myFrame.CoordinatePointTransform(myFrame.GetCoordinates(),
				   CS_CALCULATION_STANDARD,
				   temp_max);
  if(temp_min.x>temp_max.x) {
    REAL4 tx = temp_min.x ; temp_min.x=temp_max.x; temp_max.x=tx;
  }
  if(temp_min.y>temp_max.y) {
    REAL4 ty = temp_min.y ; temp_min.y=temp_max.y; temp_max.y=ty;
  }
  temp_min.z = static_cast<REAL4>(centerpt.z - arrowspan/2);
  temp_max.z = static_cast<REAL4>(centerpt.z + arrowspan/2);

  // Make shave adjustment to better mimic mmDisp output.
  if(negative_viewdir) temp_max.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
  else                 temp_min.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
  Nb_BoundingBox<REAL4> arrow_box(temp_min,temp_max);

  Nb_Vec3<REAL4> pixelsize = myFrame.GetPixelDimensions();
  temp_min.Set(static_cast<REAL4>(xmin-pixelsize.x/2),
	       static_cast<REAL4>(ymin-pixelsize.y/2),0.0f);
  temp_max.Set(static_cast<REAL4>(xmax+pixelsize.x/2),
	       static_cast<REAL4>(ymax+pixelsize.y/2),0.0f);
  myFrame.CoordinatePointTransform(myFrame.GetCoordinates(),
				   CS_CALCULATION_STANDARD,
				   temp_min);
  myFrame.CoordinatePointTransform(myFrame.GetCoordinates(),
				   CS_CALCULATION_STANDARD,
				   temp_max);
  if(temp_min.x>temp_max.x) {
    REAL4 tx = temp_min.x ; temp_min.x=temp_max.x; temp_max.x=tx;
  }
  if(temp_min.y>temp_max.y) {
    REAL4 ty = temp_min.y ; temp_min.y=temp_max.y; temp_max.y=ty;
  }
  temp_min.z = static_cast<REAL4>(centerpt.z - pixelspan/2);
  temp_max.z = static_cast<REAL4>(centerpt.z + pixelspan/2);
  // Make shave adjustment to better mimic mmDisp output.
  if(negative_viewdir) temp_max.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
  else                 temp_min.z *= static_cast<REAL4>(1+4*REAL4_EPSILON);
  Nb_BoundingBox<REAL4> pixel_box(temp_min,temp_max);

  // Compute print offset
  REAL8m pxoff=0, pyoff=0;
  if(strcmp("Landscape",page_orientation.GetStr())==0) {
    // Landscape orientation
    if(strcmp("top",page_vpos.GetStr())==0) {
      pxoff = ptmargin;
    } else if(strcmp("bottom",page_vpos.GetStr())==0) {
      pxoff = page_width - ptmargin - pheight;
    } else {
      pxoff = (page_width - pheight)/2.0;
    }
    if(strcmp("left",page_hpos.GetStr())==0) {
      pyoff = plmargin;
    } else if(strcmp("right",page_hpos.GetStr())==0) {
      pyoff = page_height - plmargin - pwidth;
    } else {
      pyoff = (page_height - pwidth)/2.0;
    }
  } else {
    // Portrait orientation
    if(strcmp("left",page_hpos.GetStr())==0) {
      pxoff = plmargin;
    } else if(strcmp("right",page_hpos.GetStr())==0) {
      pxoff = page_width - plmargin - pwidth;
    } else {
      pxoff = (page_width - pwidth)/2.0;
    }
    if(strcmp("bottom",page_vpos.GetStr())==0) {
      pyoff = ptmargin;
    } else if(strcmp("top",page_vpos.GetStr())==0) {
      pyoff = page_height - ptmargin - pheight;
    } else {
      pyoff = (page_height - pheight)/2.0;
    }
  }

  // Write PostScript
  pxoff = OC_ROUND(pxoff);
  pyoff = OC_ROUND(pyoff);
  myFrame.PSDump(channel,
		 pxoff,pyoff,
		 pwidth,pheight,
		 page_orientation.GetStr(),
		 xmin,ymin,xmax,ymax,
		 arrow_box,pixel_box,negative_viewdir,
		 mat_width,mat_color,
		 arrow_boundary_width,arrow_boundary_color);

  return TCL_OK;
}

#if (TCL_MAJOR_VERSION > 7)
struct MmdispcmdsCleanupInfo {
public:
  Tcl_Interp* interp;
  Tcl_Command bitmap_cmd_token;
} mmdispcmds_cleanup_info;
#endif

static void Mmdispcmds_Cleanup
#if (TCL_MAJOR_VERSION > 7)
(ClientData cd)
#else
(ClientData)
#endif
{
#if (TCL_MAJOR_VERSION > 7)
  MmdispcmdsCleanupInfo* info = static_cast<MmdispcmdsCleanupInfo*>(cd);
  if(info->bitmap_cmd_token != 0) {
    Tcl_DeleteCommandFromToken(info->interp,info->bitmap_cmd_token);
    info->bitmap_cmd_token = 0; // Safety
  }
#endif
  for(int i=0;i<MY_MESH_ARRAY_SIZE;i++) {
    if (myMeshArray[i]!=NULL) {
      delete myMeshArray[i];
      myMeshArray[i]=NULL;  // Safety
    }
  }
  myMesh=NULL;     // Safety
  activeMeshId=0;  // Safety
}

int Mmdispcmds_Init(Tcl_Interp *interp)
{
#define RETURN_TCL_ERROR                                                \
    Tcl_AddErrorInfo(interp, ab1("\n    (in Mmdispcmds_Init())"));      \
    return TCL_ERROR

  Oc_AutoBuf ab1,ab2;
  Tcl_HashTable *bitmapTablePtr;

  // Verify run-time and compile-time versions of required support
  // extensions match.
//  if (Tcl_PkgRequire(interp, ab1("Tcl"), ab2(TCL_VERSION), 1) == NULL) {
//    Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds "), NULL);
//    Tcl_AppendResult(interp, ab1(MMDISPCMDS_VERSION), NULL);
//    Tcl_AppendResult(interp, ab1(" needs Tcl"), NULL);
//    Tcl_AppendResult(interp, ab1(TCL_VERSION), NULL);
//    Tcl_AppendResult(interp, ab1(")"), NULL);
//  }

/*
 * Need at least Oc 1.1.1.2 to get Oc_Snprintf()
 */
    if (Tcl_PkgRequire(interp, ab1("Oc"), ab2("1.1.1.2"), 0) == NULL) {
        Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds " MMDISPCMDS_VERSION
                " needs Oc 1.1.1.2)"), NULL);
        RETURN_TCL_ERROR;
    }

/*
 * Need at least Nb 1.2.0.4 to get Nb_GetColor
 */
    if (Tcl_PkgRequire(interp, ab1("Nb"), ab2("1.2.0.4"), 0) == NULL) {
        Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds " MMDISPCMDS_VERSION
                " needs Nb 1.2.0.4)"), NULL);
        RETURN_TCL_ERROR;
    }

/*
 * Need at least Vf 1.2.0.4 to get WriteMesh routines that take Tcl_Channel
 */
    if (Tcl_PkgRequire(interp, ab1("Vf"), ab2("1.2.0.4"), 0) == NULL) {
        Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds " MMDISPCMDS_VERSION
                " needs Vf 1.2.0.4)"), NULL);
        RETURN_TCL_ERROR;
    }

  // Record compile and run-time patch levels

  // Initialize C++ data structures
  for(int id=0;id<MY_MESH_ARRAY_SIZE;id++) {
    myMeshArray[id] = new Vf_EmptyMesh();
  }
  activeMeshId=0;
  myMesh=myMeshArray[activeMeshId];
  myFrame.SetTclInterp(interp);
  myFrame.SetMesh(myMesh);

  // Register C/C++ routines with interpreter
  Oc_RegisterCommand(interp,"ChangeMesh",ChangeMesh);
  Oc_RegisterCommand(interp,"CopyMesh",CopyMesh);
  Oc_RegisterCommand(interp,"DifferenceMesh",DifferenceMesh);
  Oc_RegisterCommand(interp,"DrawFrame",DrawFrame);
  Oc_RegisterCommand(interp,"FindMeshVector",FindMeshVector);
  Oc_RegisterCommand(interp,"FreeMesh",FreeMesh);
  Oc_RegisterCommand(interp,"GetAutosamplingRates",GetAutosamplingRates);
  Oc_RegisterCommand(interp,"GetDataValueUnit",GetDataValueUnit);
  Oc_RegisterCommand(interp,"GetDataValueScaling",GetDataValueScaling);
  Oc_RegisterCommand(interp,"GetDefaultColorMapList",GetDefaultColorMapList);
  Oc_RegisterCommand(interp,"GetFrameBox",GetFrameBox);
  Oc_RegisterCommand(interp,"GetFrameRotation",GetFrameRotation);
  Oc_RegisterCommand(interp,"GetZoom",GetZoom);
  Oc_RegisterCommand(interp,"GetMeshCellSize",GetMeshCellSize);
  Oc_RegisterCommand(interp,"GetMeshCoordinates",GetMeshCoordinates);
  Oc_RegisterCommand(interp,"GetDisplayCoordinates",GetDisplayCoordinates);
  Oc_RegisterCommand(interp,"GetMeshDescription",GetMeshDescription);
  Oc_RegisterCommand(interp,"GetMeshIncrement",GetMeshIncrement);
  Oc_RegisterCommand(interp,"GetMeshName",GetMeshName);
  Oc_RegisterCommand(interp,"GetMeshStructureInfo",GetMeshStructureInfo);
  Oc_RegisterCommand(interp,"GetMeshTitle",GetMeshTitle);
  Oc_RegisterCommand(interp,"GetMeshType",GetMeshType);
  Oc_RegisterCommand(interp,"GetMeshSpatialUnitString",
		     GetMeshSpatialUnitString);
  Oc_RegisterCommand(interp,"GetMeshRange",GetMeshRange);
  Oc_RegisterCommand(interp,"GetMeshZRange",GetMeshZRange);
  Oc_RegisterCommand(interp,"GetVecColor",GetVecColor);
  Oc_RegisterCommand(interp,"GetZsliceCount",GetZsliceCount);
  Oc_RegisterCommand(interp,"GetZsliceLevels",GetZsliceLevels);
  Oc_RegisterCommand(interp,"IsRegularMesh",IsRegularMesh);
  Oc_RegisterCommand(interp,"PSWriteMesh",PSWriteMesh);
  Oc_RegisterCommand(interp,"ReportActiveMesh",ReportActiveMesh);
  Oc_RegisterCommand(interp,"SelectActiveMesh",SelectActiveMesh);
  Oc_RegisterCommand(interp,"SetDataValueScaling",SetDataValueScaling);
  Oc_RegisterCommand(interp,"SetFrameRotation",SetFrameRotation);
  Oc_RegisterCommand(interp,"SetZoom",SetZoom);
  Oc_RegisterCommand(interp,"SetMeshTitle",SetMeshTitle);
  Oc_RegisterCommand(interp,"UpdatePlotConfiguration",
                     UpdatePlotConfiguration);
  Oc_RegisterCommand(interp,"WriteMesh",WriteMesh);
  Oc_RegisterCommand(interp,"WriteMeshUsingDeprecatedVIOFormat",
		     WriteMeshUsingDeprecatedVIOFormat);
  Oc_RegisterCommand(interp,"WriteMeshAverages",WriteMeshAverages);

  bitmapTablePtr = new Tcl_HashTable;
  Tcl_InitHashTable(bitmapTablePtr, TCL_STRING_KEYS);

#if (TCL_MAJOR_VERSION > 7)
  mmdispcmds_cleanup_info.bitmap_cmd_token
    = Tcl_CreateCommand(interp, ab1("Bitmap"),BitmapCmd,
			(ClientData) bitmapTablePtr,BitmapDeleteProc);
  mmdispcmds_cleanup_info.interp = interp;
  Tcl_CreateExitHandler((Tcl_ExitProc *) Mmdispcmds_Cleanup,
			(ClientData) &mmdispcmds_cleanup_info);
#else
  Tcl_CreateCommand(interp, ab1("Bitmap"),BitmapCmd,
		    (ClientData) bitmapTablePtr,BitmapDeleteProc);
  Tcl_CreateExitHandler((Tcl_ExitProc *) Mmdispcmds_Cleanup, 0);
#endif

  if (Tcl_PkgProvide(interp, ab1("Mmdispcmds"), ab2(MMDISPCMDS_VERSION))
      != TCL_OK) {
    RETURN_TCL_ERROR;
  }
  if (Oc_InitScript(interp, "Mmdispcmds", MMDISPCMDS_VERSION) != TCL_OK) {
    RETURN_TCL_ERROR;
  }
  return TCL_OK;

#undef RETURN_TCL_ERROR

}
