# FILE: pidinfo.tcl
#
# This file must be evaluated by omfsh

########################################################################
# Support procs

# Default initializer of the accountname.  Tries to determine
# account name under which current thread is running.  This
# code is copied from the Net_Account class.
proc DefaultAccountName {} {
    global env tcl_platform
    if {[info exists tcl_platform(user)]
    	&& [string length $tcl_platform(user)]} {
        return $tcl_platform(user)
    }
    switch -- $tcl_platform(platform) {
        unix {
            if {[info exists env(USER)]} {
                return $env(USER)
            }
            if {[info exists env(LOGNAME)]} {
                return $env(LOGNAME)
            }
            return oommf
        }
        windows {
            # Ideally, we should find some reliable entry in the
            # registry.  Not yet supported.
            return oommf
        }
        default {
            return oommf
        }
    }
}

set querycount 0
proc AskServer {chan args} {
    global querycount
    puts $chan [concat query $querycount $args]
    while {[gets $chan response]>=0} {
        if {[string compare reply [lindex $response 0]]==0 \
                && [lindex $response 1]==$querycount} {
            if {[lindex $response 2]} {
                return -code error [lindex $response 3]
            }
            return [lrange $response 3 end]
        }
    }
    return -code error "Read error on server socket"
}

proc PidReport { pidreply {portsreply {}}} {
    global header
    set mypid [pid]
    if {$header} {
        set oidlength [string length "OID"]
        set pidlength [string length "PID"]
        set namelength [string length "Application"]
    } else {
        set oidlength 0
        set pidlength 0
        set namelength 0
    }
    foreach elt $pidreply {
        foreach {oid pid appname} $elt { break }
        if {$pid != $mypid} {
            set pidarr($oid) [list $pid $appname]
            if {[string length $oid]>$oidlength} {
                set oidlength [string length $oid]
            }
            if {[string length $pid]>$pidlength} {
                set pidlength [string length $pid]
            }
            if {[string length $appname]>$namelength} {
                set namelength [string length $appname]
            }
        }
    }
    if {[llength $portsreply]==0} {
        if {$header} {
            puts [format "%*s %*s %s" $oidlength OID \
                      $pidlength PID Application]
        }
        foreach oid [lsort -integer [array names pidarr]] {
            foreach {pid appname} $pidarr($oid) { break }
            puts [format "%*s %*s %s" $oidlength $oid $pidlength $pid $appname]
        }
    } else {
        if {$header} {
            puts [format "%*s %*s %*s %s" $oidlength OID \
                      $pidlength PID $namelength Application Ports]
        }
        set portlength 0
        foreach elt $portsreply {
            if {[regexp {^([0-9]+):[0-9]+$} [lindex $elt 1] dummy oid]} {
                set port [lindex $elt 2]
                lappend portarr($oid) $port
                if {[string length $port]>$portlength} {
                    set portlength [string length $port]
                }
            }
        }
        foreach oid [lsort -integer [array names pidarr]] {
            foreach {pid appname} $pidarr($oid) { break }
            puts -nonewline [format "%*s %*s %*s" $oidlength $oid \
                                 $pidlength $pid $namelength $appname]
            if {[info exists portarr($oid)]} {
                foreach port $portarr($oid) {
                    puts -nonewline [format " %*s" $portlength $port]
                }
            }
            puts {}
        }
    }
}
########################################################################

package require Oc

Oc_ForceStderrDefaultMessage
catch {wm withdraw .}

Oc_Main SetAppName pidinfo
Oc_Main SetVersion 1.2.0.4

# Remove a bunch of inapplicable default options from -help message
Oc_CommandLine Option console {} {}
Oc_CommandLine Option tk {} {}
Oc_CommandLine Option name {} {}
Oc_CommandLine Option sync {} {}
Oc_CommandLine Option id {} {}
Oc_CommandLine Option visual {} {}
Oc_CommandLine Option geometry {} {}
Oc_CommandLine Option colormap {} {}
Oc_CommandLine Option display {} {}
Oc_CommandLine Option use {} {}

set acctname [DefaultAccountName]
Oc_CommandLine Option account {
    {name {}}
} {
    global acctname;  set acctname $name
} [subst {Account (default is $acctname)}]

Oc_Option Get Net_Host port hostport  ;# Default setting
Oc_CommandLine Option hostport {
    {port {regexp {^[0-9]+$} $port}}
} {
    global hostport;  set hostport $port
} [subst {Host server port (default is $hostport)}]

Oc_CommandLine Option ports {
} {
    global showports; set showports 1
} {Display application server ports}
set showports 0

Oc_CommandLine Option noheader {
} {
    global header; set header 0
} {Don't print column header}
set header 1

Oc_CommandLine Option v {
} {
    global verbose; set verbose 1
} {Print extra info}
set verbose 0

#Oc_CommandLine Option timeout {
#    {secs {regexp {^[0-9]+$} $secs} {is timeout in seconds (default is 5)}}
#} {
#    global timeout;  set timeout [expr {$secs*1000}]
#} {Maximum time to wait for response from servers}
#set timeout 5000

Oc_CommandLine Option [Oc_CommandLine Switch] {
	{{oid list} {} {is * (default) or one or more oids}}
    } {
	global oidpat; if {[llength $oid]} {set oidpat $oid}
} {End of options; next argument is oid}
set oidpat "*"

Oc_CommandLine Parse $argv

# Only accounts on localhost currently supported
set host 127.0.0.1

if {[catch {socket $host $hostport} hostchan]} {
    puts stderr "ERROR: Unable to connect to host server\
                 on host $host, port $hostport"
    exit 1
}
fconfigure $hostchan -blocking 1 -buffering line -translation {auto crlf}
if {$verbose} {
    if {[catch {AskServer $hostchan serverpid} hostpid]} {
        set hostpid "???"
    }
}
if {[catch {AskServer $hostchan lookup $acctname} acctport]} {
    puts stderr "Account lookup failure: $acctport"
    close $hostchan
    exit 1
}
close $hostchan

set acctpid 0
if {$verbose} {
    puts "-----------------------"
    puts "        Host: $host"
    puts "    Host pid: $hostpid"
    if {$showports} { puts "   Host port: $hostport" }
    puts "     Account: $acctname"
}


if {[catch {socket $host $acctport} acctchan]} {
    puts stderr "ERROR: Unable to connect to account server\
                 on host $host, port $acctport"
    exit 1
}
fconfigure $acctchan -blocking 1 -buffering line -translation {auto crlf}
if {$verbose} {
    if {[catch {AskServer $acctchan serverpid} acctpid]} {
        set acctpid "???"
    }
    puts " Account pid: $acctpid"
    if {$showports} { puts "Account port: $acctport" }
    puts "-----------------------"
}
if {[catch {AskServer $acctchan getpids $oidpat -appname} pidreply]} {
    puts stderr "Pid lookup failure: $pidreply"
    close $acctchan
    exit 1
}
set portsreply {}
if {$showports} {
    if {[catch {AskServer $acctchan threads} portsreply]} {
        puts stderr "Ports lookup failure: $portsreply"
        close $acctchan
        exit 1
    }
}
close $acctchan

PidReport $pidreply $portsreply

exit
