# FILE: procs.tcl
#
# Misc. Tcl support code for Oxs
#

# Proc to read a file.  This is a wrapper for use by safe
# interpreters, and also provides image conversion facility.
# The second argument should be one of "binary", "auto" or
# "image".  The first two govern the translation mode of the
# opened file.  Selecting "image" causes the input to be
# converted to mimic the body of a PPM P3 text file---the first
# 3 whitespace separated values are width, height, and maxvalue.
# After this is an array or pixels, r g b.  The difference
# compared to a straight P3 file is that the leading "P3" is
# removed, and there are no comments.  The output can be treated
# as a list and used directly.
#
# The return value is a string representing the read data.
#
proc Oxs_ReadFile { file {translation auto} } {
    if {![string match image $translation]} {
	# Standard file read
	set chanid [open $file r]
	if {[string match binary $translation]} {
	    fconfigure $chanid -translation binary
	}
	fconfigure $chanid -buffering full -buffersize 50000
	set filestr [read $chanid]
	close $chanid
	return $filestr
    } else {
	# Read and convert to simple text representation
	set chanid [open $file r]
	fconfigure $chanid -translation binary \
		-buffering full -buffersize 250000
	set type_check [read $chanid 2]
	if {![string match P3 $type_check]} {
	    # Input file is not PPM P3 (ASCII) type.
	    # Send through converter.
	    close $chanid
	    set var {}
	    if {[catch {
		Oc_Application CommandLine any2ppm -noinfo -o - $file
	    } cmdLine]} {
		return -code error "Can't find any2ppm program to\
			convert	image file $file to PPM P3 format."
	    }
	    set cmdLine [linsert $cmdLine 0 |]
	    # Ignore output set the stderr.  Without this redirection,
	    # any output to stderr causes Tcl to raise an error on
	    # closing the channel.  In particular, some debugging
	    # libraries print out version information at program start
	    # on stderr.
	    lappend cmdLine "2>" \
                    [[Oc_Config RunPlatform] GetValue path_device_null]
	    set chanid [open $cmdLine r]
	    fconfigure $chanid -translation binary \
		    -buffering full -buffersize 250000
	    set type_check [read $chanid 2]
	    if {![string match P3 $type_check]} {
		return -code error "Conversion of image file $file to\
			PPM P3 format failed."
	    }
	}
	set image [read $chanid] ;# Note that leading P3 already stripped
	close $chanid
    
	# Remove comments, if any
	regsub -all -- "#\[^\n]*\n" $image "\n" image
    
	return $image
    }
}

# Restricted glob, for use by slightly unsafe interpreters.
# This version does not allow access to directories other
# than the current working directory.
proc Oxs_RGlob { args } {
    if {[llength $args]<1} {
	return -code error "wrong # args: should be\
		\"Glob ?switches? name ?name ...?\""
    }
    set types {}
    for {set index 0} {$index<[llength $args]} {incr index} {
	set option [lindex $args $index]
	if {![string match -* $option]} { break }
	if {[string match -- $option]} {
	    incr index
	    break
	}
	switch -exact -- $option {
	    -types {
		if {$index+1>=[llength $args]} {
		    return -code error "missing argument to \"-types\""
		}
		incr index
		set types [lindex $args $index]
	    }
	    default {
		return -code error "bad option \"$option\":\
			must be -types or --"
	    }
	}
    }

    set matchlist {}
    set filelist [glob -nocomplain -types $types *]
    foreach file $filelist {
	foreach pat [lrange $args $index end] {
	    if {[string match $pat $file]} {
		lappend matchlist $file
		break
	    }
	}
    }
    return $matchlist
}
